
"""

**Overview**
------------

This implements a model of the `nback task <https://en.wikipedia.org/wiki/N-back#Neurobiology_of_n-back_task>`_
described in `Beukers et al. (2022) <https://psyarxiv.com/jtw5p>`_.  The model uses a simple implementation of episodic
memory (EM, as a form of content-retrieval memory) to store previous stimuli along with the temporal context in which
they occurred, and a feedforward neural network (FFN) to evaluate whether the current stimulus is a match to the n'th
preceding stimulus (n-back level) retrieved from EM.

The model is an example of proposed interactions between working memory (subserved by neocortical structures) and
episodic memory (subserved by hippocampus, and possibly cerebellum) in the performance of tasks demanding of sequential
processing and control, along the lines of models emerging from machine learning that augment the use of recurrent
neural networks (e.g., long short-term memory mechanisms; LSTMs) for active memory and control, with an external memory
capable of rapid storage and content-based retrieval, such as the
Neural Turing Machine (NTN; `Graves et al., 2016 <https://arxiv.org/abs/1410.5401>`_),
Episodic Planning Networks (EPN; `Ritter et al., 2020 <https://arxiv.org/abs/2006.03662>`_), and
Emergent Symbols through Binding Networks (ESBN; `Webb et al., 2021 <https://arxiv.org/abs/2012.14601>`_).

The script contains methods to construct, train, and run the model, and analyze the results of its execution:

* `construct_model <nback.construct_model>`:
  takes as arguments parameters used to construct the model;  for convenience, defaults are defined below,
  (under "Construction parameters")

* `train_network <nback.train_network>`:
  takes as arguments the feedforward neural network Composition (FFN_COMPOSITION) and number of epochs to train.
  Note: learning_rate is set at construction (can specify using LEARNING_RATE under "Training parameters" below).

* `run_model <nback.run_model>`:
  takes as arguments the drift rate in the temporal context vector to be applied on each trial,
  and the number of trials to execute, as well as reporting and animation specifications
  (see "Execution parameters").

* `analyze_results <nback.analyze_results>`:
  takes as arguments the results of executing the model, and optionally a number of trials and nback_level to analyze;
  returns d-prime statistics and plots results for different conditions at each nback_level executed.


**The Model**
-------------

The model is comprised of two `Compositions <Composition>`: an outer one that contains the full model (`nback_model
<nback_model_composition>`), and an `AutodiffComposition`, nested within nback_model, that implements the feedforward
neural network (`ffn <nback_ffn_composition>`) (see red box in the figure below).  Both of these are constructed in
the `construct_model <nback.construct_model>` function (see `below <nback_methods_reference>`).

.. _nback_Fig:

.. figure:: _static/N-Back_Model_movie.gif
   :align: left
   :alt: N-Back Model Animation

.. _nback_model_composition:

*nback_model Composition*
~~~~~~~~~~~~~~~~~~~~~~~~~

This is comprised of three input Mechanisms, and the nested `ffn <nback_ffn_composition>` `Composition`.

.. _nback_ffn_composition:

*FFN Composition*
~~~~~~~~~~~~~~~~~

The temporal context is provided by a randomly drifting high dimensional vector that maintains a constant norm (i.e.,
drifts on a sphere).  The FFN is trained, given an n-back level of *n*, to identify when the current stimulus matches
one stored in EM with a temporal context vector that differs by an amount corresponding to *n* time steps of drift.
During n-back performance, the model encodes the current stimulus and temporal context, retrieves an item from EM
that matches the current stimulus, weighted by the similarity of its temporal context vector (i.e., most recent), and
then uses the FFN to evaluate whether it is an n-back match.  The model responds "match" if the FFN detects a match;
otherwise, it either responds "non-match" or, with a fixed probability (hazard rate), it uses the current stimulus
and temporal context to retrieve another sample from EM and repeat the evaluation.

The ffn Composition is trained using the train_network() method


**Construction and Execution**
------------------------------

.. _nback_settings:

*Settings*
~~~~~~~~~~

The default parameters are ones that have been fit to empirical data concerning human performance
(taken from `Kane et al., 2007 <https://psycnet.apa.org/record/2007-06096-010?doi=1>`_).

See "Settings for running the script" to specify whether the model is trained and/or executed when the script is run,
and whether a graphic display of the network is generated when it is constructed.

.. _nback_stimuli:

*Stimuli*
~~~~~~~~~

Sequences of stimuli are constructed either using `SweetPea <https://sites.google.com/view/sweetpea-ai?pli=1>`_
(using the script in stim/SweetPea) or replicate those used in the study by `Kane et al.,
2007 <https://psycnet.apa.org/record/2007-06096-010?doi=1>`_ (from stimulus files in stim/Kane_et_al).

    .. note::
       Use of SweetPea for stimulus generation requires it be installed::
       >> pip install sweetpea

.. _nback_training:

*Training*
~~~~~~~~~~

MORE HERE

.. _nback_execution:

*Execution*
~~~~~~~~~~~

MORE HERE

.. _nback_methods_reference:

**Methods Reference**
---------------------

COMMENT:
TODO:
    - from Andre
             - network architecture;  in particular, size of hidden layer and projection patterns to and from it
                - the stim+context input vector (length 90) projects to a hidden layer (length 80);
                - the task input vector (length 2) projects to a different hidden layer (length 80);
                - those two hidden layers project (over fixed, nonlearnable, one-one-projections?) to a third hidden layer (length 80) that simply sums them;
                - the third hidden layer projects to the length 2 output layer;
                - a softmax is taken over the output layer to determine the response.
                - fix: were biases trained?
          - training:
              - learning rate: 0.001; epoch: 1 trial per epoch of training
              - fix: state_dict with weights (still needed)
          - get empirical stimulus sequences (still needed)
          - put nback script (with pointer to latest version on PNL) in nback-paper repo
    - train_network() and run_model(): refactor to take inputs and trial_types, and training_set, respectively
    - fix: get rid of objective_mechanism (see "VERSION *WITHOUT* ObjectiveMechanism" under control(...)
    - fix: warnings on run
    - fix: remove num_nback_levels from contruct_model
    - complete documentation in BeukersNbackModel.rst
    - validate against nback-paper results
    - after validation:
        - try with STIM_SIZE = NUM_STIMS rather than 20 (as in nback-paper)
        - refactor generate_stim_sequence() to use actual empirical stimulus sequences
        - replace get_input_sequence and _get_training_inputs with generators passed to nback_model.run() and ffn.learn
        - build version that *can* maintain in WM, and uses EVC to decide which would be easier:
           maintenance in WM vs. storage/retrieval from EM (and the fit to Jarrod's data)
COMMENT

"""

import os
import random
import time
import timeit
import numpy as np
from typing import Union
from enum import Enum, IntEnum
from pathlib import Path

from graph_scheduler import *

from psyneulink import *

# Settings for running script:
CONSTRUCT_MODEL = True # THIS MUST BE SET TO True to run the script
DISPLAY_MODEL = False # True = show visual graphic of model
TRAIN_FFN = True  # True => train the FFN (WM)
TEST_FFN = True  # True => test the FFN on training stimuli (WM)
RUN_MODEL = True  # True => test the model on sample stimulus sequences
ANALYZE_RESULTS = True # True => output analysis of results of run
REPORT_OUTPUT = ReportOutput.OFF       # Sets console output during run
REPORT_PROGRESS = ReportProgress.OFF   # Sets console progress bar during run
ANIMATE = False # {UNIT:EXECUTION_SET} # Specifies whether to generate animation of execution

#region ========================================= PARAMETERS ===========================================================

# Fixed parameters:
MAX_NBACK_LEVELS = 3
NUM_STIM = 8 # number of different stimuli in stimulus set -  QUESTION: WHY ISN"T THIS EQUAL TO STIM_SIZE OR VICE VERSA?
FFN_TRANSFER_FUNCTION = ReLU

# Constructor parameters:  (values are from nback-paper)


STIM_SIZE = 20 # length of stimulus vector
CONTEXT_SIZE = 25 # length of temporal context vector
TASK_SIZE = 10 # length of task specification vector
FFN_INPUT_SIZE = (STIM_SIZE + CONTEXT_SIZE) * 2 # length of full input vector
H1_SIZE = FFN_INPUT_SIZE # dimension of stimulus hidden units in ff
H2_SIZE = 80
NBACK_LEVELS = [2,3] # Currently restricted to these
NUM_NBACK_LEVELS = len(NBACK_LEVELS)
CONTEXT_DRIFT_NOISE = 0.0  # noise used by DriftOnASphereIntegrator (function of Context mech)
RANDOM_WEIGHTS_INITIALIZATION=RandomMatrix(center=0.0, range=0.1)  # Matrix spec used to initialize all Projections
DROPOUT_PROB = 0.05
RETRIEVAL_SOFTMAX_TEMP = 1 / 8 # express as gain # precision of retrieval process
RETRIEVAL_HAZARD_RATE = 0.04 # rate of re=sampling of em following non-match determination in a pass through ffn
RETRIEVAL_STIM_WEIGHT = .05 # weighting of stimulus field in retrieval from em
RETRIEVAL_CONTEXT_WEIGHT = 1 - RETRIEVAL_STIM_WEIGHT # weighting of context field in retrieval from em
# DECISION_SOFTMAX_TEMP=1

# Training parameters:
NUM_TRAINING_SETS_PER_EPOCH = 1
MINIBATCH_SIZE=None
NUM_EPOCHS = 6250 # 12500 # 20000  # nback-paper: 400,000 @ one trial per epoch = 6,250 @ 64 trials per epoch
FOILS_ALLOWED_BEFORE = False
LEARNING_RATE=0.001  # nback-paper: .001

# Execution parameters:
CONTEXT_DRIFT_RATE=.1 # drift rate used for DriftOnASphereIntegrator (function of Context mech) on each trial
NUM_TRIALS = 48 # number of stimuli presented in a trial sequence

# Names of Compositions and Mechanisms:
NBACK_MODEL = "nback Model"
FFN_COMPOSITION = "WORKING MEMORY (fnn)"
FFN_INPUT = "CURRENT INPUT LAYER"
FFN_TASK = "CURRENT TASK LAYER"
FFN_TASK_EMBED = "TASK EMBEDDING LAYER"
FFN_H1 = "H1 LAYER"
FFN_ADD_LAYER = "ADD LAYER"
FFN_DROPOUT = "DROPOUT LAYER"
FFN_H2 = "H2 LAYER"
FFN_OUTPUT = "OUTPUT LAYER"
MODEL_STIMULUS_INPUT ='STIM'
MODEL_CONTEXT_INPUT = 'CONTEXT'
MODEL_TASK_INPUT = "TASK"
CONCATENATE_FFN_INPUT = "CONCATENATE INPUT"
EM = "EPISODIC MEMORY (dict)"
DECISION = "DECISION"
CONTROLLER = "READ/WRITE CONTROLLER"


class TrialTypes(Enum):
    """Trial types explicitly assigned and counter-balanced in _get_run_inputs()
    In notation below, "A" is always current stimulus.
    Foils are only explicitly assigned to items immediately following nback item,
        or before if **foils_allowed_before** is specified in _get_training_inputs()
    Subseq designated below as "not explicitly assigned" may still appear in the overall stimulus seq,
        either within the subseq through random assignment,
        and/or through cross-subseq relationships that are not controlled in this design.
    """
    MATCH_NO_FOIL = 'match'             # ABA (2-back) or ABCA (3-back); not explicitly assigned: ABBA
    MATCH_WITH_FOIL = 'stim_lure'       # AAA (2-back) or AABA (3-back); not explicitly assigned: ABAA or AAAA
    NO_MATCH_NO_FOIL = 'non_lure'       # BBA (2-back) or BCDA (3-back); not explicitly assigned: BBCA, BCCA or BBBA
    NO_MATCH_WITH_FOIL = 'context_lure' # BAA (2-back) or BACA (3-back); not explicitly assigned: BCAA or BAAA


num_trial_types = len(TrialTypes)


class Stimuli(Enum):
    SWEETPEA = 'sweetpea'
    KANE_STIMULI = 'kane'

#endregion

#region ===================================== MODEL CONSTRUCTION =======================================================

def construct_model(stim_size:int = STIM_SIZE,
                    context_size:int = CONTEXT_SIZE,
                    ffn_input_size:int = FFN_INPUT_SIZE,
                    task_size = TASK_SIZE,
                    h1_size:int = H1_SIZE,
                    h2_size:int = H2_SIZE,
                    num_nback_levels:int = NUM_NBACK_LEVELS,
                    context_drift_noise:float = CONTEXT_DRIFT_NOISE,
                    retrievel_softmax_temp:float = RETRIEVAL_SOFTMAX_TEMP,
                    retrieval_hazard_rate:float = RETRIEVAL_HAZARD_RATE,
                    retrieval_stimulus_weight:float = RETRIEVAL_STIM_WEIGHT,
                    retrieval_context_weight:float = RETRIEVAL_CONTEXT_WEIGHT,
                    )->Composition:
    """**Construct nback_model**

    Arguments
    ---------
    stim_size: int
      dimensionality of stimulus vector
    context_size: int
      dimensionality of context vector
    ffn_input_size: int
      dimensionality of input to ffn (current stimulus and context + retrieved stimulus and context)
    task_size: int
      dimensionality of task embedding layer
    h1_size: int
      dimensionality of  input embedding layer
    h2_size: int
      dimensionality of hidden layer
    num_nback_levels: int
      number of nback_levels to implement
    context_drift_noise: float
      rate of temporal context drift
    retrievel_softmax_temp: float
      temperature of softmax on retrieval from episodic memory
    retrieval_hazard_rate: float
      rate at which episodic memory is sampled if a match is not found
    retrieval_stimulus_weight: float
      weighting on stimulus component for retrieval of vectors stored in `episodic memory <EpisodicMemoryMechanism>`
    retrieval_context_weight: float
      weighting on context component for retrieval of vectors stored in `episodic memory <EpisodicMemoryMechanism>`

    Returns
    -------
    Composition implementing nback model
    """

    print(f"constructing '{FFN_COMPOSITION}'...")

    # FEED FORWARD NETWORK -----------------------------------------
    #     inputs: encoding of current stimulus and context, retrieved stimulus and retrieved context,
    #     task input:  one hot
    #     output: match [1,0] or non-match [0,1]
    # Must be trained to detect match for specified task (1-back, 2-back, etc.)
    stim_context_input = TransferMechanism(name=FFN_INPUT,
                                           input_shapes=ffn_input_size)
    task_input = ProcessingMechanism(name=FFN_TASK,
                                     input_shapes=task_size)
    task_embedding = ProcessingMechanism(name=FFN_TASK,
                                         input_shapes=h1_size)
    h1 = ProcessingMechanism(name=FFN_H1,
                             input_shapes=h1_size,
                             function=FFN_TRANSFER_FUNCTION)
    add_layer = ProcessingMechanism(name=FFN_ADD_LAYER,
                                    input_shapes=h1_size)
    dropout = ProcessingMechanism(name=FFN_DROPOUT,
                                  input_shapes=h1_size,
                                  function=Dropout(p=DROPOUT_PROB))
    h2 = ProcessingMechanism(name=FFN_H2,
                             input_shapes=h2_size,
                             function=FFN_TRANSFER_FUNCTION)
    output = ProcessingMechanism(name=FFN_OUTPUT,
                                 input_shapes=2,
                                 function = Linear
                                 # function=ReLU
                                 )
    PASS_THROUGH = MappingProjection(matrix = IDENTITY_MATRIX, exclude_in_autodiff=True),
    input_pway = Pathway([stim_context_input, RANDOM_WEIGHTS_INITIALIZATION, h1, IDENTITY_MATRIX, add_layer])
    task_pway = Pathway([task_input, RANDOM_WEIGHTS_INITIALIZATION, task_embedding, IDENTITY_MATRIX, add_layer])
    output_pway = Pathway([add_layer, IDENTITY_MATRIX, dropout,
                           RANDOM_WEIGHTS_INITIALIZATION, h2, RANDOM_WEIGHTS_INITIALIZATION, output])

    ffn = AutodiffComposition(pathways = [input_pway, task_pway, output_pway],
                              name=FFN_COMPOSITION,
                              learning_rate=LEARNING_RATE,
                              optimizer_type='adam',
                              # optimizer_type='sgd',
                              loss_spec=Loss.CROSS_ENTROPY
                              # loss_spec=Loss.MSE
                              )

    # FULL MODEL (Outer Composition, including input, EM and control Mechanisms) ------------------------

    print(f"constructing '{NBACK_MODEL}'...")

    # Stimulus Encoding: takes stim_size vector as input
    stim = TransferMechanism(name=MODEL_STIMULUS_INPUT, input_shapes=stim_size)

    # Context Encoding: takes scalar as drift step for current trial
    context = ProcessingMechanism(name=MODEL_CONTEXT_INPUT,
                                  function=DriftOnASphereIntegrator(
                                      initializer=np.random.random(context_size - 1),
                                      noise=context_drift_noise,
                                      dimension=context_size))

    # Task: task one-hot indicating n-back (1, 2, 3 etc.) - must correspond to what ffn has been trained to do
    task = ProcessingMechanism(name=MODEL_TASK_INPUT,
                               input_shapes=task_size)

    # Episodic Memory:
    #    - entries: stimulus (field[0]) and context (field[1]); randomly initialized
    #    - uses Softmax to retrieve best matching input, subject to weighting of stimulus and context by STIM_WEIGHT
    em = EpisodicMemoryMechanism(name=EM,
                                 input_ports=[{NAME:"STIMULUS_FIELD",
                                               INPUT_SHAPES:stim_size},
                                              {NAME:"CONTEXT_FIELD",
                                               INPUT_SHAPES:context_size}],
                                 function=ContentAddressableMemory(
                                     initializer=[[[0] * stim_size, [0] * context_size]],
                                     distance_field_weights=[retrieval_stimulus_weight,
                                                             retrieval_context_weight],
                                     # equidistant_entries_select=NEWEST,
                                     selection_function=SoftMax(output=MAX_INDICATOR,
                                                                gain=retrievel_softmax_temp)))

    # Input to FFN
    concat_input = ProcessingMechanism(name=CONCATENATE_FFN_INPUT,
                                       input_ports=[stim, context,
                                                    em.output_ports["RETRIEVED_STIMULUS_FIELD"],
                                                    em.output_ports["RETRIEVED_CONTEXT_FIELD"]],
                                       function=Concatenate)

    decision = TransferMechanism(name=DECISION,
                                 input_shapes=2,
                                 function=SoftMax(output=MAX_INDICATOR))

    # Control Mechanism
    #  Ensures current stimulus and context are only encoded in EM once (at beginning of trial)
    #    by controlling the storage_prob parameter of em:
    #      - if outcome of decision signifies a match or hazard rate is realized:
    #        - set  EM[store_prob]=1 (as prep encoding stimulus in EM on next trial)
    #        - this also serves to terminate trial (see nback_model.termination_processing condition)
    #      - if outcome of decision signifies a non-match and hazard rate is not realized:
    #        - set  EM[store_prob]=0 (as prep for another retrieval from EM without storage)
    #        - continue trial
    control = ControlMechanism(name=CONTROLLER,
                               default_variable=[[1]],  # Ensure EM[store_prob]=1 at beginning of first trial
                               # ---------
                               # VERSION *WITH* ObjectiveMechanism:
                               objective_mechanism=ObjectiveMechanism(name="OBJECTIVE MECHANISM",
                                                                      monitor=decision,
                                                                      # Outcome=1 if match, else 0
                                                                      function=lambda x: int(x[0][0]>x[0][1])),
                               # Set ControlSignal for EM[store_prob]
                               #   to 1 if match or hazard rate is realized (i.e., store stimulus and end trial)
                               #   else 0 (i.e., don't store stimulus and continue retrieving)
                               function=lambda outcome: int(bool(outcome)
                                                            or (np.random.random() < retrieval_hazard_rate)),
                               # ---------
                               # # VERSION *WITHOUT* ObjectiveMechanism:
                               # monitor_for_control=decision,
                               # # Set Evaluate outcome and set ControlSignal for EM[store_prob]
                               # #   - outcome is received from decision as one hot in the form: [[match, no-match]]
                               # function=lambda outcome: int(int(outcome[0][1]>outcome[0][0])
                               #                              or (np.random.random() > retrieval_hazard_rate)),
                               # ---------
                               control=(STORAGE_PROB, em))

    nback_model = Composition(name=NBACK_MODEL,
                              # nodes=[stim, context, task, ffn, em, logit, decision, control],
                              nodes=[stim, context, task, ffn, em, concat_input, decision, control],
                              # Terminate trial if value of control is still 1 after first pass through execution
                              termination_processing={TimeScale.TRIAL: And(Condition(lambda: control.value),
                                                                           AfterPass(0, TimeScale.TRIAL))},
                              )
    nback_model.add_projection(MappingProjection(), stim, em.input_ports["STIMULUS_FIELD"])
    nback_model.add_projection(MappingProjection(), context, em.input_ports["CONTEXT_FIELD"])
    nback_model.add_projection(MappingProjection(), task, task_input)
    nback_model.add_projection(MappingProjection(), output, decision, IDENTITY_MATRIX)
    nback_model.add_projection(MappingProjection(), concat_input, stim_context_input)

    if DISPLAY_MODEL:
        nback_model.show_graph(
            # show_cim=True,
            # show_node_structure=ALL,
            # show_dimensions=True
        )

    print(f'full model constructed')
    return nback_model
#endregion

#region =====================================STIMULUS GENERATION =======================================================

def _get_stim_set(stim_size=STIM_SIZE, num_stim=NUM_STIM):
    """Construct an array of unique stimuli for use in an experiment, used by train_network() and run_model()"""
    # For now, use one-hots
    return np.eye(stim_size)[0:num_stim]

def _get_task_input(nback_level):
    """Construct input to task Mechanism for a given nback_level, used by train_network() and run_model()"""
    task_input = [0] * TASK_SIZE
    task_input[nback_level - NBACK_LEVELS[0]] = 1
    return task_input

def _get_training_inputs(network:AutodiffComposition,
                         num_training_sets_per_epoch:int=NUM_TRAINING_SETS_PER_EPOCH,
                         num_epochs:int=1,
                         nback_levels:int=NBACK_LEVELS,
                         foils_allowed_before:bool=True,
                         return_generator:bool=True
                         )->(dict, list, int):
    """Construct set of training stimuli used by ffn.learn() in train_network()
    Construct one example of each condition for each stimulus and each nback_level:
        MATCH_NO_FOIL (match):  stim_current = stim_retrieved  and context_current = context_retrieved
        MATCH_WITH_FOIL (stim_lure):  stim_current = stim_retrieved  and context_current != context_retrieved
        NO_MATCH_WITH_FOIL (context_lure):  stim_current != stim_retrieved  and context_current == context_retrieved
        NO_MATCH_NO_FOIL (non_lure):  stim_current != stim_retrieved  and context_current != context_retrieved
    Arguments
    ---------
    network: AutodiffComposition
        network to be trained.
    num_training_sets_per_epoch: int : default 1
        number of complete sets of training stimuli to be trained in a single epoch (i.e., for a single weight update);
        used to determine batch_size (= num_training_sets_per_epoch * number of trials in a training set)
    num_epochs: int : default 1
        number of epochs of training to be executed (passed to `learn <AutodiffComposition.learn>`)
    nback_levels: list[int] : default NBACK_LEVELS
        list of n-back levels for which to generate training sets;
        nback_levels themselves must be specified in the global NBACK_LEVELS
    foils_allowed_before: bool : default True
        only allows foils to occur after the target (e.g., for 2-back: XBAA and not ABXA)
    return_generator: bool : True
        return generator rather than explicit list of training stimuli
    Return
    ------
    (training_set: dict,  conditions: list[TrialType], batch_size: int)
    """

    assert is_iterable(nback_levels) and all([0<i<=MAX_NBACK_LEVELS for i in nback_levels])
    stimuli = _get_stim_set()
    num_stim = len(stimuli)
    if num_stim != NUM_STIM:
        warnings.warn(f"Number of stimuli used for training set ({num_stim}) is not same "
                      f"as specified in NUM_STIM ({NUM_STIM}); unexpexpected results may occur.")
    context_fct = DriftOnASphereIntegrator(initializer=np.random.random(CONTEXT_SIZE - 1),
                                           noise=CONTEXT_DRIFT_NOISE,
                                           dimension=CONTEXT_SIZE)

    def trial_gen():
        for ep in range(num_epochs):
            stim_array = []
            target = []
            current_task = []
            for n in range(num_training_sets_per_epoch):
                for nback_level in nback_levels:
                    # Construct one hot encoding for nback level
                    task_input = _get_task_input(nback_level)
                    # Construct an example of each trial type for each stimulus
                    for i in range(num_stim):
                        contexts = []
                        # Get current stimulus and a randomly selected distractor
                        stims = list(stimuli.copy())
                        # Get stim, and remove from stims so distractor can be picked randomly from remaining ones
                        current_stim = stims.pop(i)
                        # Pick distractor randomly from stimuli remaining in set
                        distractor_stim = stims[np.random.randint(0,len(stims))]

                        # IF foils_allowed_before IS True:
                        #    total_contexts = nback+2 contexts:  [0]=potential foil; [1]=nback, [2+nback]=current
                        # IF foils_allowed_before IS False:
                        #    total_contexts = nback+1 contexts:  [0]=nback, [1+nback]=current
                        total_contexts = nback_level + 1 + int(foils_allowed_before)
                        for j in range(total_contexts):
                            contexts.append(context_fct(CONTEXT_DRIFT_RATE))
                        # Get current context as last in list
                        current_context = contexts.pop(-1)
                        # Get nback context as either first or second in list, based on foils_allowed_before
                        nback_context = contexts.pop(int(foils_allowed_before))
                        # Choose distractor foil randomly from remaining contexts
                        # (note:  if foils_allowed_before = False, only those after target remain in list)
                        distractor_context = contexts[np.random.randint(0,len(contexts))]

                        # Assign retrieved stimulus and context accordingly to trial_type
                        for trial_type in TrialTypes:
                            # Assign current stimulus as retrieved stimulus for MATCH_ trials
                            if trial_type in {TrialTypes.MATCH_NO_FOIL, TrialTypes.MATCH_WITH_FOIL}:
                                stim_retrieved = current_stim
                            # Assign distractor stimulus as retrieved stimulus for NON_MATCH_ trials
                            else:
                                stim_retrieved = distractor_stim
                            # Assign nback context as retrieved context for _NO_FOIL trials
                            if trial_type in {TrialTypes.MATCH_NO_FOIL,TrialTypes.NO_MATCH_NO_FOIL}:
                                context_retrieved = nback_context
                            # Assign distractor context as retrieved context for _WITH_FOIL trials
                            else:
                                context_retrieved = distractor_context
                            # Assign target
                            if trial_type in {TrialTypes.MATCH_NO_FOIL, TrialTypes.MATCH_WITH_FOIL}:
                                target.append([1,0])
                            else:
                                target.append([0,1])
                            current_task.append([task_input])
                            stim_array.append([np.array(current_stim.tolist()
                                                        + current_context.tolist()
                                                        + stim_retrieved.tolist()
                                                        + context_retrieved.tolist())])
            inputs = {network.nodes[FFN_INPUT]: stim_array}
            targets = {network.nodes[FFN_OUTPUT]: target}

            training_set = {INPUTS: inputs,
                            TARGETS: targets,
                            # EPOCHS: num_epochs}
                            EPOCHS: 1}

            yield training_set

    # Get training_set
    if return_generator:
        training_set = trial_gen()
    else:
        training_set = next(trial_gen())

    # Get sequence of conditions
    conditions = []
    for n in range(num_training_sets_per_epoch):
        for l in range(len(nback_levels)):
            for i in range(num_stim):
                for trial_type in TrialTypes:
                    conditions.append(trial_type.name)

    batch_size = num_stim * len(TrialTypes) * len(nback_levels) * num_training_sets_per_epoch

    return training_set, conditions, batch_size,

def _get_run_inputs(model,
                    nback_level:int,
                    context_drift_rate:float=CONTEXT_DRIFT_RATE,
                    num_trials:int=NUM_TRIALS,
                    mini_blocks:bool=True,
                    inputs_source:Stimuli=None)->(dict,list):
    """Construct set of stimulus inputs for run_model(), balancing across four conditions.
    Trial_type assignments:
      - trial_types are assigned to subseqs of nback_level+1 stimuli that are concatenated to form the full trial seq
      - trial_type subseqs are constructed in get_stim_subseq_for_trial_type(), by randomly picking a target stimulus,
          and then assigning the preceding stimuli in the subseq to conform the trial_type
      - the balancing of trial_types is enforced *only* for the last stimulus in each set;
          the others are inferred and may not be fully balanced across conditions
          (depending on number of stimuli, this may be impossible).
    Mini_blocks:
      - if True (default) trials are sequenced in mini-blocks each of which contains one set of trials
          for each trial_type; order of trial_type subseq within each mini_block is randomized across them;
          number of trials in a mini-block = nback_level+1 * num_trial_types; trials not assigned to
          mini_blocks (i.e., modulus of num_trial % (num_mini_blocks * mini_block_size) are assigned random
          stimuli and trial_type is inferred posthoc).
      - if False, sampling of trial_types is balanced,
          but order of presentation is randomized over the entire sequence
    Returns
    -------
    dict with inputs to each input node of model for each trial and array with corresponding trial_type_assignments
    """

    def generate_stim_sequence(nback_level, num_trials):
        assert nback_level in {2,3} # At present, only 2- and 3-back levels are supported

        stim_set = _get_stim_set()

        def get_stim_subseq_for_trial_type(trial_type):
            """Return stimulus seq (as indices into stim_set) for the specified trial_type."""
            subseq_size = nback_level + 1
            subseq = [None] * subseq_size
            curr_stim = subseq[nback_level] = random.choice(np.arange(len(stim_set)))
            other_stims = np.setdiff1d(np.arange(len(stim_set)),curr_stim).tolist()
            trial_type = list(TrialTypes)[trial_type]

            if trial_type == TrialTypes.MATCH_NO_FOIL:           # ABA (2-back) or ABCA (3-back)
                subseq[0] = curr_stim  # Assign nback stim to match
                # Assign remaining items in sequence to any stimuli other than curr_stim
                subseq[1:nback_level] = random.sample(other_stims, nback_level - 1)
            elif trial_type == TrialTypes.MATCH_WITH_FOIL:        # AAA (2-back) or AABA (3-back)
                subseq[0] = curr_stim  # Assign nback stim to match current stim
                subseq[1] = curr_stim  # Assign curr_stim to stim next to nback as foil
                # Assign any remaining items in sequence to any stimuli other than curr_stim
                subseq[2:nback_level] = random.sample(other_stims, nback_level - 2)
            elif trial_type == TrialTypes.NO_MATCH_NO_FOIL:       # ABB (2-back) or BCDA (3-back)
                # Assign remaining items in sequence to any stimuli than curr_stim
                subseq[0:nback_level] = random.sample(other_stims, nback_level)
            elif trial_type == TrialTypes.NO_MATCH_WITH_FOIL:     # BAA (2-back) or BACA (3-back)
                # Assign remaining items in sequence to any stimuli than curr_stim
                subseq[1] = curr_stim  # Assign curr_stim to stim next to nback as foil
                subseq[0:1] = random.sample(other_stims, 1)
                subseq[2:nback_level] = random.sample(other_stims, nback_level - 2)
            assert None not in subseq, "Failed to assign all stims for subseq in get_stim_subseq_for_trial_type."
            return subseq

        def get_trial_type_for_stim(subseq):
            if subseq[-1] == subseq[0] and not subseq[-1] in subseq[1:-1]:
                return TrialTypes.MATCH_NO_FOIL.value
            elif subseq[-1] == subseq[0] and subseq[-1] in subseq[0:-1]:
                return TrialTypes.MATCH_WITH_FOIL.value
            elif subseq[-1] not in subseq[0:-1]:
                return TrialTypes.NO_MATCH_NO_FOIL.value
            elif subseq[-1] != subseq[0] and subseq[-1] in subseq[0:-1]:
                # Note: for 3back, this includes: BAXA, BXAA, and BAAA
                return TrialTypes.NO_MATCH_WITH_FOIL.value

        subseq_size = nback_level + 1
        num_sub_seqs = int(num_trials / num_trial_types)
        extra_trials = num_trials % num_trial_types

        # Construct seq of mini-blocks (subseqs) each containing one sample of each trial_type in a random order
        #    note: this is done over number of mini-blocks that fit into num_trials;
        #          remaining trials are randomly assigned trial_types below

        num_mini_blocks = int(num_trials / (num_trial_types * (nback_level + 1)))
        mini_block_size = subseq_size * num_trial_types # Number of trials in a mini_block
        seq_of_trial_type_subseqs = []
        # Generate randomly ordered trial_type assignments for subseqs in each mini_block
        for i in range(num_mini_blocks):
            seq_of_trial_type_subseqs.extend(random.sample(range(num_trial_types), num_trial_types))
        if not mini_blocks:
            # Randomize the order of trial types across the entire sequence:
            random.shuffle(seq_of_trial_type_subseqs)

        if extra_trials:  # Warn if conditions can't be fully balanced
            warnings.warn(f"Number of trials ({num_trials}) precludes fully balancing across all five trial types")

        stim_seq = [None] * num_trials
        trial_type_seq = [None] * num_trials
        # Construct actual stimulus sequence by getting stimuli for each subseq, up to num_sub_seqs
        #   note: the trial type only applies (and a trial_type is only assigned) to the last trial of each subsequence;
        #         trial_type of preceding ones set below on the full sequence of stimuli is assigned
        # stim_seq.append(get_stim_seq_for_trial_type(i) for i in seq_of_trial_type_subseqs) # <- CONDENSED VERSION
        for i, trial_type in enumerate(seq_of_trial_type_subseqs):  # <- LOOP VERSION
            idx = i * subseq_size
            # Get seq of stimuli for subseq of specified trial_type
            stim_seq[idx:idx + nback_level + 1] = get_stim_subseq_for_trial_type(trial_type)
            # Assign trial_type to last stim in subseq (since it was constructed specifically for that trial_type)
            trial_type_seq[idx + nback_level] = list(TrialTypes)[trial_type].value
        # Pad remainder to get to num_trials with randomly selected stimuli
        stim_seq.extend(random.sample(range(num_trial_types),extra_trials))
        # Infer trial_types for all remaining stimuli (which should currently be marked as None)
        for i in range(subseq_size,num_trials,subseq_size):
            for j in range(i,i + nback_level):
                assert trial_type_seq[j] is None, f"trial_type should still be None for trial {j}."
                trial_type_seq[j] = get_trial_type_for_stim(stim_seq[i - subseq_size:i])
                assert True

        trial_type_counts = [None] * num_trial_types
        for i in range(num_trial_types):
            trial_type_counts[i] = trial_type_seq.count(i)

        return(stim_seq, trial_type_seq)

    def get_input_sequence(nback_level, num_trials=NUM_TRIALS, inputs_source=None):
        """Construct sequence of stimulus and trial_type indices"""
        # Use SweetPea if specified
        if inputs_source == Stimuli.SWEETPEA:
            if nback_level == 2:
                from stim.SweetPea.sweetpea_script import create_two_back
                Kane_stimuli = {stim:idx for idx, stim in enumerate(['B', 'F', 'H', 'K', 'M', 'Q', 'R', 'X'])}
                Kane_trial_types = {'1/1/0': TrialTypes.MATCH_NO_FOIL,
                                    '1/2/0': TrialTypes.MATCH_WITH_FOIL,
                                    '2/1/0': TrialTypes.NO_MATCH_NO_FOIL,
                                    '2/2/0': TrialTypes.NO_MATCH_WITH_FOIL}
                stim_dict = create_two_back()
                assert True
                stim_seq = [Kane_stimuli[i.upper()] for i in stim_dict[0]['letter']]
                trial_type_seq = [Kane_trial_types[i] if i else None for i in stim_dict[0]['condi']]
                assert True
            else:
                raise Exception(f"Use of SweetPea currently restricted to nback_level = 2")
        elif inputs_source == Stimuli.KANE_STIMULI:
            assert False, "KANE STIMULI NOT YET SUPPORTED AS INPUTS"
        # Else, use local algorithm
        else:
            stim_seq, trial_type_seq = generate_stim_sequence(nback_level, num_trials)
            # Return list of corresponding stimulus input vectors

        input_set = [_get_stim_set()[i] for i in stim_seq]
        return input_set, trial_type_seq

    input_set, trial_type_seq = get_input_sequence(nback_level, num_trials, inputs_source=inputs_source)
    return {model.nodes[MODEL_STIMULUS_INPUT]: input_set,
            model.nodes[MODEL_CONTEXT_INPUT]: [[context_drift_rate]] * num_trials,
            model.nodes[MODEL_TASK_INPUT]: [_get_task_input(nback_level)] * num_trials}, \
           trial_type_seq
#endregion

#region ================================== MODEL EXECUTION METHODS =====================================================

def train_network(network:AutodiffComposition,
                  training_set:dict=None,
                  minibatch_size:int=MINIBATCH_SIZE,
                  learning_rate:float=LEARNING_RATE,
                  num_epochs:int=NUM_EPOCHS,
                  save_weights_to:Union[Path,str,None]=None
                  )->Path:
    """**Train feedforward network** on example stimulus sequences for each condition.

    Arguments
    ---------
    network: AutodiffComposition
        network to be trained;  this must be an `AutodiffComposition`.
    training_set: dict : default _get_training_inputs()
        inputs (see `Composition_Input_Dictionary`), including targets (`Composition_Target_Inputs`)
        to use for training;  these are constructed in a call to _get_training_inputs() if not specified here.
    minibatch_size: int : default MINIBATCH_SIZE
        number of inputs that will be presented within a single training epoch
        (i.e. over which weight changes are aggregated and applied);  if it is not specified and MINIBATCH_SIZE=None,
        it is determined by the batch_size for an epoch returned in the call to _get_training_inputs().
    learning_rate: float : default LEARNING_RATE
        learning_rate to use for training;  this overrides the value of `learning_rate
        <AutodiffComposition.learning_rate>` specified in construction of the network.  If None is specified
         here, either the value specified at construction, or the default for `AutodiffComposition
         <AutodiffComposition.learning_rate>` is used.
    num_epochs: int : default NUM_EPOCHS
        number of training epochs (i.e., sets of minibatchs) to execute during training.
    save_weights_to: Path : 'results' subdirectory of current working directory
        location to store weights at end of training.

    Returns
    -------
    Path containing saved weights for matrices of feedforward Projections in network.
    """
    print(f"\nconstructing training set for '{network.name}'...")
    if training_set is None:
        training_set, conditions, batch_size = \
            _get_training_inputs(network=network,
                                 num_training_sets_per_epoch=NUM_TRAINING_SETS_PER_EPOCH,
                                 foils_allowed_before=FOILS_ALLOWED_BEFORE,
                                 num_epochs=num_epochs,
                                 nback_levels=NBACK_LEVELS)
    else:
        batch_size = len(training_set[MODEL_STIMULUS_INPUT])

    minibatch_size = minibatch_size or batch_size
    print(f'num training stimuli per training set: {minibatch_size//NUM_TRAINING_SETS_PER_EPOCH}')
    print(f'num training sets per epoch: {NUM_TRAINING_SETS_PER_EPOCH}')
    print(f'total num training stimuli per epoch: {minibatch_size}')
    print(f'num epochs (weight updates): {num_epochs}')
    print(f'total num trials: {num_epochs*minibatch_size}')
    print(f"\ntraining '{network.name}' (started at {time.localtime()[3]%12}:{'{:02}'.format(time.localtime()[4])})...")
    start_time = timeit.default_timer()
    network.learn(inputs=training_set,
                  minibatch_size=minibatch_size,
                  num_trials=minibatch_size,
                  # report_output=REPORT_OUTPUT,
                  # report_progress=REPORT_PROGRESS,
                  # report_progress=ReportProgress.ON,
                  learning_rate=learning_rate,
                  # execution_mode=ExecutionMode.LLVMRun
                  # execution_mode=ExecutionMode.Python
                  execution_mode=ExecutionMode.PyTorch
                  )
    stop_time = timeit.default_timer()
    print(f"training of '{network.name}' done")
    training_time = stop_time - start_time
    if training_time <= 60:
        training_time_str = f'{int(training_time)} seconds'
    else:
        training_time_str = f'{int(training_time/60)} minutes {int(training_time%60)} seconds'
    print(f'training time: {training_time_str} for {num_epochs} epochs')
    path = network.save(filename=save_weights_to, directory="results_original")
    print(f'saved weights to: {save_weights_to}')
    return path
    # print(f'saved weights sample: {network.nodes[FFN_H1].path_afferents[0].matrix.base[0][:3]}...')
    # network.load(path)
    # print(f'loaded weights sample: {network.nodes[FFN_H1].path_afferents[0].matrix.base[0][:3]}...')

def network_test(network:AutodiffComposition,
                 load_weights_from:Union[Path,str,None]=None,
                 nback_levels=NBACK_LEVELS,
                 )->(dict,list,list,list,list,list):

    print(f"constructing test set for '{network.name}'...")
    test_set, conditions, set_size = _get_training_inputs(network=network,
                                                          num_epochs=1,
                                                          nback_levels=NBACK_LEVELS,
                                                          foils_allowed_before=FOILS_ALLOWED_BEFORE,
                                                          return_generator=False)
    print(f'total num trials: {set_size}')

    inputs = [test_set[INPUTS][network.nodes['CURRENT INPUT LAYER']][i] for i in range(set_size)]
    # cxt_distances = [Distance(metric=COSINE)([inputs[i][2],inputs[i][3]]) for i in range(set_size)]
    current_stim_idx = slice(0,STIM_SIZE)
    current_context_idx = slice(current_stim_idx.stop, current_stim_idx.stop + CONTEXT_SIZE)
    retrieved_stim_idx = slice(current_context_idx.stop, current_context_idx.stop + STIM_SIZE)
    retrieved_context_idx = slice(retrieved_stim_idx.stop, retrieved_stim_idx.stop + CONTEXT_SIZE)
    stim_distances = [Distance(metric=COSINE)([inputs[i][0][current_stim_idx],
                                              inputs[i][0][retrieved_stim_idx]]) for i in range(set_size)]
    cxt_distances = [Distance(metric=COSINE)([inputs[i][0][current_context_idx],
                                              inputs[i][0][retrieved_context_idx]]) for i in range(set_size)]
    targets = list(test_set[TARGETS].values())[0]
    trial_type_stats = []

    num_items_per_nback_level = int(set_size / NUM_NBACK_LEVELS)
    for i in range(NUM_NBACK_LEVELS):
        start = i * num_items_per_nback_level
        stop = start + num_items_per_nback_level
        stimulus_distances_for_level = np.array(stim_distances[start:stop])
        context_distances_for_level = np.array(cxt_distances[start:stop])
        conditions_for_level = np.array(conditions[start:stop])
        for trial_type in TrialTypes:
            trial_type_stats.append(
                (f'{NBACK_LEVELS[i]}-back', trial_type.name, trial_type.value,
                 context_distances_for_level[np.where(conditions_for_level==trial_type.name)].mean(),
                 context_distances_for_level[np.where(conditions_for_level==trial_type.name)].std()))

    # FIX: COMMENTED OUT TO TEST TRAINING LOSS
    if load_weights_from:
        print(f"nback_model loading '{FFN_COMPOSITION}' weights from {load_weights_from}...")
        network.load(filename=load_weights_from)

    network.run(inputs=test_set[INPUTS], report_progress=ReportProgress.ON)

    if ANALYZE_RESULTS:
        coded_responses, stats = analyze_results([network.results,conditions], test=True)
        import torch
        cross_entropy_loss = \
            [network.loss_function(torch.Tensor(output[0]),torch.Tensor(np.array(target))).detach().numpy().tolist()
             for output, target in zip(network.results, targets)]
    coded_responses_flat = []
    for nback_level in nback_levels:
        coded_responses_flat.extend(coded_responses[nback_level])
    return inputs, cxt_distances, targets, conditions, network.results, coded_responses_flat, cross_entropy_loss, \
           trial_type_stats, stats

def run_model(model,
              load_weights_from:Union[Path,str,None]=None,
              context_drift_rate:float=CONTEXT_DRIFT_RATE,
              num_trials:int=NUM_TRIALS,
              inputs_source:Stimuli=None,
              report_output:ReportOutput=REPORT_OUTPUT,
              report_progress:ReportProgress=REPORT_PROGRESS,
              animate:Union[dict,bool]=ANIMATE,
              save_results_to:Union[Path,str,None]=None
              )->list:
    """**Run model** for all nback levels with a specified context drift rate and number of trials

    Arguments
    --------
    load_weights_from:  Path
        specifies file from which to load pre-trained weights for matrices of FFN_COMPOSITION.
    context_drift_rate: float
        specifies drift rate as input to CONTEXT_INPUT, used by DriftOnASphere function of FFN_CONTEXT_INPUT.
    num_trials: int
        number of trials (stimuli) to run.
    report_output: ReportOutput
        specifies whether to report results during execution of run (see `Report_Output` for additional details).
    report_progress: ReportProgress
        specifies whether to report progress of execution during run (see `Report_Progress` for additional details).
    animate: dict or bool
        specifies whether to generate animation of execution (see `ShowGraph_Animation` for additional details).
    save_results_to: Path
        specifies location to save results of the run along with trial_type_sequences for each nback level;
        if None, those are returned by call but not saved.
    """
    ffn = model.nodes[FFN_COMPOSITION]
    em = model.nodes[EM]
    if load_weights_from:
        print(f"nback_model loading '{FFN_COMPOSITION}' weights from {load_weights_from}...")
        ffn.load(filename=load_weights_from)
    print(f"'{model.name}' executing...")
    trial_type_seqs = [None] * NUM_NBACK_LEVELS
    start_time = timeit.default_timer()
    for i, nback_level in enumerate(NBACK_LEVELS):
        # Reset episodic memory for new task using first entry (original initializer)
        em.function.reset(em.memory[0])
        inputs, trial_type_seqs[i] = _get_run_inputs(model=model,
                                                     nback_level=nback_level,
                                                     context_drift_rate=context_drift_rate,
                                                     num_trials=num_trials,
                                                     inputs_source=inputs_source)
        model.run(inputs=inputs,
                  report_output=report_output,
                  report_progress=report_progress,
                  animate=animate
                  )
    # print("Number of entries in EM: ", len(model.nodes[EM].memory))
    stop_time = timeit.default_timer()
    assert len(model.nodes[EM].memory) == NUM_TRIALS + 1 # extra one is for initializer
    if REPORT_PROGRESS == ReportProgress.ON:
        print('\n')
    print(f"'{model.name}' done: {len(model.results)} trials executed")
    execution_time = stop_time - start_time
    if execution_time <= 60:
        execution_time_str = f'{int(execution_time)} seconds'
    else:
        execution_time_str = f'{int(execution_time/60)} minutes {int(execution_time%60)} seconds'
    print(f'execution time: {execution_time_str}')
    results = np.array([model.results, trial_type_seqs], dtype=object)
    if save_results_to:
        np.save(save_results_to, results)
    # print(f'results: \n{model.results}')
    if ANALYZE_RESULTS:
        coded_responses, stats = analyze_results(results, test=False)
    return results, coded_responses, stats
#endregion

#region ================================= MODEL PERFORMANCE ANALYSIS ===================================================

def analyze_results(results:list,
                    nback_levels:list=NBACK_LEVELS,
                    test:bool=False
                    )->(dict,dict):
    """**Analyze and plot results** of executed model.

    Arguments
    --------
    results: ndarray
        results returned from `run_model <nback.run_model>`.
    nback_levels: list : default NBACK_LEVELS
        list of nback levels executed in run() or test()
    test: bool : default False
        if True, analyze results for running ffn on set of stimuli used for training
        else, anaylze results of running full model using experimental sequence of stimuli
    """
    responses_and_trial_types = [None] * len(nback_levels)
    stats = np.zeros((len(nback_levels),num_trial_types))
    MATCH = 'match'
    NON_MATCH = 'non-match'
    num_trials = int(len(results[0]) / len(nback_levels))

    # FOR TEST
    if test:
        print(f"\n\nTest results (of ffn on training set):")
        for i, nback_level in enumerate(nback_levels):
            # conditions = results[1][i]
            conditions = results[1]
            # Code responses for given nback_level as 1 (match) or 0 (non-match)
            responses_for_nback_level = [r[0] for r in results[0][i * num_trials:i * num_trials + num_trials]]
            responses_for_nback_level = [MATCH if r[0] > r[1] else NON_MATCH for r in responses_for_nback_level]
            responses_and_trial_types[i] = list(zip(responses_for_nback_level, conditions))
            for j, trial_type in enumerate(TrialTypes):
                relevant_data = [[response, condition] for response, condition in zip(responses_for_nback_level, conditions)
                                 if condition == trial_type.name]
                # Report % matches in each condition (should be 1.0 for MATCH trials and 0.0 for NON-MATCH trials
                stats[i][j] = [d[0] for d in relevant_data
                                        if d[0] is not None].count(MATCH) / (len(relevant_data))
            print(f"nback level {nback_level}:")
            for j, performance in enumerate(stats[i]):
                print(f"\t{list(TrialTypes)[j].name}: {performance:.1f}")

    # FOR RUN
    else:
        print(f"\n\nTest results (running full model on experimental sequence):")
        for i, nback_level in enumerate(nback_levels):
            conditions = results[1][i]
            # Code responses for given nback_level as 1 (match) or 0 (non-match)
            responses_for_nback_level = [r[0] for r in results[0][i * num_trials:i * num_trials + num_trials]]
            responses_for_nback_level = [MATCH if r[0] > r[1] else NON_MATCH for r in responses_for_nback_level]
            responses_and_trial_types[i] = list(zip(responses_for_nback_level, conditions))
            for j, trial_type in enumerate(TrialTypes):
                relevant_data = [[response, condition] for response, condition in zip(responses_for_nback_level,
                                                                                      conditions)
                                 if condition == trial_type.value]
                # Report % matches in each condition (should be 1.0 for MATCH trials and 0.0 for NON-MATCH trials
                stats[i][j] = [d[0] for d in relevant_data
                                        if d[0] is not None].count(MATCH) / (len(relevant_data))
            print(f"nback level {nback_level}:")
            for j, performance in enumerate(stats[i]):
                print(f"\t{list(TrialTypes)[j].name}: {performance:.1f}")

    data_dict = {k:v for k,v in zip(nback_levels, responses_and_trial_types)}
    stats_dict = {}
    for i, nback_level in enumerate(nback_levels):
        stats_dict.update({nback_level: {trial_type.name:stat for trial_type,stat in zip(TrialTypes, stats[i])}})

    return data_dict, stats_dict

def _compute_dprime(hit_rate, fa_rate):
    """returns dprime and sensitivity
    """
    def clamp(n, minn, maxn):
        return max(min(maxn, n), minn)
    # hit_rate = clamp(hit_rate, 0.01, 0.99)
    # fa_rate = clamp(fa_rate, 0.01, 0.99)

    dl = np.log(hit_rate * (1 - fa_rate) / ((1 - hit_rate) * fa_rate))
    c = 0.5 * np.log((1 - hit_rate) * (1 - fa_rate) / (hit_rate * fa_rate))
    return dl, c

def _plot_results(response_and_trial_types, stats):

    # SCORE IN NBACK-PAPER IS RETURNED BY THIS METHOD:
    # def run_EMexp(neps,tsteps,argsD):
    #   score = np.zeros([2,4,tsteps,neps])
    #   for ep in range(neps):
    #     for nback in nbackL:
    #       for seq_int,tstep in itertools.product(range(4),np.arange(5,tsteps)):
    #         print(ep,tstep)
    #         stim,ctxt,ytarget = generate_trial(nback,tstep,stype=seq_int)
    #         yhat = run_model_trial(stim,ctxt,nback-2,argsD)
    #         [nback-2,seq_int,tstep,ep] = int(yhat==ytarget)
    #   print(score.shape) # nback,seqtype,tsteps,epoch
    #   acc = score.mean((2,3))
    #   return acc,score

    import matplotlib as plt
    hits_stderr = np.concatenate((score.mean(2).std(-1) / np.sqrt(neps))[:,(0,1)])
    correj_stderr = np.concatenate((score.mean(2).std(-1) / np.sqrt(neps))[:,(2,3)])
    d,s = _compute_dprime(
      np.concatenate(score.mean(2)[:,(0,1)]),
      np.concatenate(score.mean(2)[:,(2,3)])
    )
    print(d.shape,s.shape)
    dprime_stderr = d.std(-1) / np.sqrt(neps)
    bias_stderr = s.std(-1) / np.sqrt(neps)
    #%%
    # 2back-target, 2back-lure, 3back-target, 3back-lure
    hits = np.concatenate(acc[:,(0,1)])
    correj = np.concatenate(acc[:,(2,3)])
    dprime = np.zeros(4)
    bias = np.zeros(4)
    for i in range(4):
        d,s = _compute_dprime(hits[i], 1 - correj[i])
        dprime[i]=d
        bias[i]=s

    #%%
    f,axar = plt.subplots(2,2,figsize=(15,8))
    axar=axar.reshape(-1)
    cL = ['blue','darkblue','lightgreen','forestgreen']
    labL = ['2b,ctrl','2b,lure','3b,ctrl','3b,lure']

    # correct reject
    ax = axar[0]
    ax.set_title('correct rejection')
    ax.bar(range(4),correj,color=cL,yerr=correj_stderr)

    # hits
    ax = axar[1]
    ax.set_title('hits')
    ax.bar(range(4),hits,color=cL,yerr=hits_stderr)

    #
    ax = axar[2]
    ax.set_title('dprime')
    ax.bar(range(4),dprime,color=cL,yerr=dprime_stderr)

    #
    ax = axar[3]
    ax.set_title('bias')
    ax.bar(range(4),bias,color=cL,yerr=bias_stderr)

    ##
    for ax in axar[:2]:
        ax.set_xticks(np.arange(4))
        ax.set_xticklabels(labL)
        ax.set_ylim(0,1)

    plt.savefig('figures/EMmetrics-%s-t%i.jpg' % (mtag,tstamp))
    plt.savefig('figures/EMmetrics_yerr-%s-t%i.svg' % (mtag,tstamp))
#endregion


#region ===================================== SCRIPT EXECUTION =========================================================
# Construct, train and/or run model based on settings at top of script

# Only execute if called from command line (i.e., not on import)
if __name__ == '__main__':
    if CONSTRUCT_MODEL:
        nback_model = construct_model()

    if TRAIN_FFN:
        weights_filename = f'results/ffn.wts_nep_{NUM_EPOCHS}_lr_{str(LEARNING_RATE).split(".")[1]}.pnl'
        weights_path = Path('/'.join([os.getcwd(), weights_filename]))
        saved_weights = train_network(nback_model.nodes[FFN_COMPOSITION],
                                      save_weights_to=weights_path
                                      )

    if TEST_FFN:
        try:
            weights_path
        except:
            weights_filename = f'results/ffn.wts_nep_{NUM_EPOCHS}_lr_{str(LEARNING_RATE).split(".")[1]}.pnl'
            weights_path = Path('/'.join([os.getcwd(), weights_filename]))

        inputs, cxt_distances, targets, conditions, results, coded_responses, ce_loss, \
        trial_type_stats, stats = \
            network_test(nback_model.nodes[FFN_COMPOSITION],
                         load_weights_from = weights_path
                         )
        headings = ['condition', 'inputs', 'target', 'context distance', 'results', 'coded response', 'ce loss']
        results = (headings,
                   list(zip(conditions, inputs, targets, cxt_distances, results, coded_responses, ce_loss)),
                   trial_type_stats,
                   stats)

        # SAVE RESULTS in CSV FORMAT
        import csv
        threshold = .005

        high_loss = [list(x) for x in [results[1][i] for i in range(64)] if x[6] > threshold]
        for i in range(len(high_loss)):
            high_loss[i][6] = '{:.4f}'.format(high_loss[i][6])
        high_loss.insert(0,headings)
        file = open('high_loss.csv', 'w+', newline ='')
        with file:
            write = csv.writer(file)
            write.writerows(high_loss)
        file.close()

        low_loss = [list(x) for x in [results[1][i] for i in range(64)] if x[6] <= threshold]
        for i in range(len(low_loss)):
            low_loss[i][6] = '{:.4f}'.format(low_loss[i][6])
        low_loss.insert(0,headings)
        file = open('low_loss.csv', 'w+', newline ='')
        with file:
            write = csv.writer(file)
            write.writerows(low_loss)
        file.close()

        full_results = [list(x) for x in [results[1][i] for i in range(64)]]
        for i in range(len(full_results)):
            full_results[i][6] = '{:.4f}'.format(full_results[i][6])
        full_results.insert(0,headings)
        file = open('full_results.csv', 'w+', newline ='')
        with file:
            write = csv.writer(file)
            write.writerows(full_results)
        file.close()

    if RUN_MODEL:
        results_path = Path('/'.join([os.getcwd(), f'results/nback.results_nep_{NUM_EPOCHS}_lr'
                                                   f'_{str(LEARNING_RATE).split(".")[1]}.pnl']))
        try:
            weights_path
        except:
            weights_filename = f'results/ffn.wts_nep_{NUM_EPOCHS}_lr_{str(LEARNING_RATE).split(".")[1]}.pnl'
            weights_path = Path('/'.join([os.getcwd(), weights_filename]))
        results = run_model(nback_model,
                            load_weights_from = weights_path,
                            save_results_to= results_path
                            # inputs_source=Stimuli.SWEETPEA,
                            )
    # if ANALYZE_RESULTS:
    #     coded_responses, stats = analyze_results(results,
    #                                              nback_levels=NBACK_LEVELS)
#endregion



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM bindings **************************************************************

import ctypes
import enum
import functools
import gc
import inspect
import numpy as np
import time
from math import ceil, log2
from psyneulink._typing import Set
import weakref

from llvmlite import ir

from . import codegen
from .builder_context import *
from .builder_context import _all_modules, _convert_llvm_ir_to_ctype, _convert_llvm_ir_to_dtype
from .debug import debug_env
from .execution import *
from .execution import _tupleize
from .jit_engine import *
from .warnings import *


__all__ = ['LLVMBuilderContext', 'ExecutionMode']

class ExecutionMode(enum.Flag):
    """Specify execution a `Composition` in interpreted or one of ithe compiled modes.
    These are used to specify the **execution_mode** argument of a Composition's `execute <Composition.execute>`,
    `run <Composition.run>`, and `learn <Composition.learn>` methods.  See `Compiled Modes
    <Composition_Compilation_Modes>` under `Compilation <Composition_Compilation>` for additional details concerning
    use of each mode by a Composition.

    Attributes
    ----------

    Python
      Execute using the Python interpreter;  this is the default mode.

    LLVM
      compile and run Composition `Nodes <Composition_Nodes>` and `Projections <Projection>` individually.

    LLVMExec
      compile and run each `TRIAL <TimeScale.TRIAL>` individually.

    LLVMRun
      compile and run multiple `TRIAL <TimeScale.TRIAL>`\\s.

    Auto
      progressively attempt LLVMRun, LLVMexec. LLVM and then Python.

    PyTorch
      execute the `AutodiffComposition` `learn <AutodiffComposition.learn>` method using PyTorch, and its
      `run <AutodiffComposition.run>` method using the Python interpreter.

      .. warning::
         For clarity, this mode should only be used when executing an `AutodiffComposition`; using it
         with a standard `Composition` is possible, but it will **not** have the expected effect of executing
         its `learn <Composition.learn>` method using PyTorch.

    PTX
      compile and run Composition `Nodes <Composition_Nodes>` and `Projections <Projection>` using CUDA for GPU.

    PTXRun
      compile and run multiple `TRIAL <TimeScale.TRIAL>`\\s using CUDA for GPU.
   """

    Python   = 0
    PyTorch = enum.auto()
    LLVM     = enum.auto()
    PTX      = enum.auto()
    _Run      = enum.auto()
    _Exec     = enum.auto()
    _Fallback = enum.auto()

    Auto = _Fallback | _Run | _Exec | LLVM
    LLVMRun = LLVM | _Run
    LLVMExec = LLVM | _Exec
    PTXRun = PTX | _Run
    COMPILED = ~ (Python | PyTorch)


_binary_generation = 0


def _compiled_modules() -> Set[ir.Module]:
    return set().union(*(e.compiled_modules for e in _get_engines()))


def _staged_modules() -> Set[ir.Module]:
    return set().union(*(e.staged_modules for e in _get_engines()))


def _llvm_build(target_generation=_binary_generation + 1):
    global _binary_generation
    if target_generation <= _binary_generation:
        if "compile" in debug_env:
            print("SKIPPING COMPILATION: {} -> {}".format(_binary_generation, target_generation))
        return

    if "compile" in debug_env:
        print("STAGING GENERATION: {} -> {}".format(_binary_generation, target_generation))

    for e in _get_engines():
        e.stage_compilation(_modules)
    _modules.clear()

    # update binary generation
    _binary_generation = target_generation


class LLVMBinaryFunction:
    def __init__(self, name: str, *, ctype_ptr_args:tuple=(), dynamic_size_args:tuple=()):
        self.name = name

        self.__c_func = None
        self.__cuda_kernel = None

        # Make sure builder context is initialized
        LLVMBuilderContext.get_current()

        # Compile any pending modules
        _llvm_build(LLVMBuilderContext._llvm_generation)

        # Function signature
        # We could skip compilation if the function is in _compiled_models,
        # but that happens rarely
        f = _find_llvm_function(self.name, _compiled_modules() | _staged_modules())

        # Create ctype function instance
        start = time.perf_counter()
        return_type = _convert_llvm_ir_to_ctype(f.return_value.type)

        self.np_arg_dtypes = [_convert_llvm_ir_to_dtype(getattr(a.type, "pointee", a.type)) for a in f.args]

        args = [_convert_llvm_ir_to_ctype(a.type) for a in f.args]

        # '_type_' special attribute stores pointee type for pointers
        # https://docs.python.org/3/library/ctypes.html#ctypes._Pointer._type_
        self.byref_arg_types = [a._type_ if hasattr(a, "contents") else None for a in args]

        for i, arg in enumerate(self.np_arg_dtypes):
            if i not in ctype_ptr_args and self.byref_arg_types[i] is not None:
                if i in dynamic_size_args:
                    args[i] = np.ctypeslib.ndpointer(dtype=arg.base, ndim=len(arg.shape) + 1, flags='C_CONTIGUOUS')
                else:
                    args[i] = np.ctypeslib.ndpointer(dtype=arg.base, shape=arg.shape, flags='C_CONTIGUOUS')

        middle = time.perf_counter()
        self.__c_func_type = ctypes.CFUNCTYPE(return_type, *args)
        finish = time.perf_counter()

        if "time_stat" in debug_env:
            print("Time to create ctype function '{}': {} ({} to create types)".format(
                  name, finish - start, middle - start))

    @property
    def c_func(self):
        if self.__c_func is None:
            # This assumes there are potential staged modules.
            # The engine had to be instantiated to have staged modules,
            # so it's safe to access it directly
            _cpu_engine.compile_staged()
            ptr = _cpu_engine._engine.get_function_address(self.name)
            self.__c_func = self.__c_func_type(ptr)
        return self.__c_func

    def __call__(self, *args, **kwargs):
        return self.c_func(*args, **kwargs)

    @property
    def _cuda_kernel(self):
        if self.__cuda_kernel is None:
            _ptx_engine.compile_staged()
            self.__cuda_kernel = _ptx_engine.get_kernel(self.name)
        return self.__cuda_kernel

    def cuda_max_block_size(self, override):
        if override is not None:
            return override

        kernel = self._cuda_kernel
        device = jit_engine.pycuda.autoinit.device

        if kernel.shared_size_bytes > 0:
            # we use shared memory, prefer big blocks.
            # Limited by reg usage
            rounded_regs = 2 ** ceil(log2(kernel.num_regs))
            block_size = device.get_attribute(jit_engine.pycuda.driver.device_attribute.MAX_REGISTERS_PER_BLOCK) // rounded_regs
        else:
            # Use smallest possible blocks
            block_size = device.get_attribute(jit_engine.pycuda.driver.device_attribute.WARP_SIZE)

        block_size = min(device.get_attribute(jit_engine.pycuda.driver.device_attribute.MAX_THREADS_PER_BLOCK), block_size)
        if "stat" in debug_env:
            print("kernel '", self.name, "' registers:", kernel.num_regs)
            print("kernel '", self.name, "' local memory size:", kernel.local_size_bytes)
            print("kernel '", self.name, "' shared memory size:", kernel.shared_size_bytes)
            print("kernel '", self.name, "' selected block size:", block_size)

        return block_size

    def cuda_call(self, *args, threads=1, block_size=None):
        block_size = self.cuda_max_block_size(block_size)
        grid = ((threads + block_size - 1) // block_size, 1)
        ktime = self._cuda_kernel(*args, np.int32(threads), time_kernel="time_stat" in debug_env,
                                  block=(block_size, 1, 1), grid=grid)
        if "time_stat" in debug_env:
            print("Time to run kernel '{}' using {} threads: {}".format(
                self.name, threads, ktime))

    def cuda_wrap_call(self, *args, **kwargs):
        wrap_args = (jit_engine.pycuda.driver.InOut(a) if isinstance(a, np.ndarray) else a for a in args)
        self.cuda_call(*wrap_args, **kwargs)

    def np_buffer_for_arg(self, arg_num, *, extra_dimensions=(), fill_value=np.nan):

        out_base = self.np_arg_dtypes[arg_num].base
        out_shape = extra_dimensions + self.np_arg_dtypes[arg_num].shape

        # fill the buffer with NaN poison
        return np.full(out_shape, fill_value, dtype=out_base)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def from_obj(obj, *, tags:frozenset=frozenset(), ctype_ptr_args:tuple=(), dynamic_size_args:tuple=()):
        name = LLVMBuilderContext.get_current().gen_llvm_function(obj, tags=tags).name
        return LLVMBinaryFunction.get(name, ctype_ptr_args=ctype_ptr_args, dynamic_size_args=dynamic_size_args)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def get(name: str, *, ctype_ptr_args:tuple=(), dynamic_size_args:tuple=()):
        return LLVMBinaryFunction(name, ctype_ptr_args=ctype_ptr_args, dynamic_size_args=dynamic_size_args)


_cpu_engine = None
_ptx_engine = None

def _get_engines():
    global _cpu_engine
    if _cpu_engine is None:
        _cpu_engine = cpu_jit_engine()

    global _ptx_engine
    if ptx_enabled:
        if _ptx_engine is None:
            _ptx_engine = ptx_jit_engine()
        return [_cpu_engine, _ptx_engine]

    return [_cpu_engine]



def cleanup(check_leaks:bool=False):
    global _cpu_engine
    _cpu_engine = None
    global _ptx_engine
    _ptx_engine = None

    _modules.clear()
    _all_modules.clear()

    LLVMBinaryFunction.get.cache_clear()
    LLVMBinaryFunction.from_obj.cache_clear()

    if check_leaks and LLVMBuilderContext.is_active():
        old_context = LLVMBuilderContext.get_current()

        LLVMBuilderContext.clear_global()

        # check that WeakKeyDictionary is not keeping any references
        # Try first without calling the GC
        c = weakref.WeakSet(old_context._cache.keys())
        if len(c) > 0:
            gc.collect()

        assert len(c) == 0, list(c)
    else:
        LLVMBuilderContext.clear_global()

        # If not checking for leaks, there might be active compositions that
        # cache pointers to binary functions. Accessing those pointers would
        # cause segfault.
        # Extract the set of associated compositions. Both to avoid duplicate
        # clears for executions that belong to the same composition, and to
        # avoid modifying the container that is iterated over.
        for c in {e._composition for e in CompExecution.active_executions}:
            c._compilation_data.execution.values.clear()
            c._compilation_data.execution.history.clear()

        # The set of active executions should be empty
        for e in CompExecution.active_executions:
            assert any(inspect.isframe(r) for r in gc.get_referrers(e))

        CompExecution.active_executions.clear()



import psyneulink as pnl
import pytest

import optuna

from psyneulink.core.components.functions.nonstateful.fitfunctions import (
    PECOptimizationFunction,
)

import numpy as np
import pandas as pd


# Define function to generate a counterbalanced trial sequence with a specified switch trial frequency
def generate_trial_sequence(n, frequency, seed: int = None):

    # Compute trial number
    nTotalTrials = n
    switchFrequency = frequency

    nSwitchTrials = int(nTotalTrials * switchFrequency)
    nRepeatTrials = int(nTotalTrials - nSwitchTrials)

    # Determine task transitions
    transitions = [1] * nSwitchTrials + [0] * nRepeatTrials
    rng = np.random.RandomState(seed)
    order = rng.permutation(list(range(nTotalTrials)))
    transitions[:] = [transitions[i] for i in order]

    # Determine stimuli with 50% congruent trials
    stimuli = (
        [[1, 1]] * int(nSwitchTrials / 4)
        + [[1, -1]] * int(nSwitchTrials / 4)
        + [[-1, -1]] * int(nSwitchTrials / 4)
        + [[-1, 1]] * int(nSwitchTrials / 4)
        + [[1, 1]] * int(nRepeatTrials / 4)
        + [[1, -1]] * int(nRepeatTrials / 4)
        + [[-1, -1]] * int(nRepeatTrials / 4)
        + [[-1, 1]] * int(nRepeatTrials / 4)
    )
    stimuli[:] = [stimuli[i] for i in order]

    # stimuli[:] = [[1, 1]] * nTotalTrials

    # Determine cue-stimulus intervals
    CSI = (
        [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nSwitchTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
        + [1200] * int(nRepeatTrials / 8)
    )
    CSI[:] = [CSI[i] for i in order]

    # Set the task order
    tasks = [[1, 0]] * (nTotalTrials + 1)
    for i in list(range(nTotalTrials)):
        if transitions[i] == 0:
            tasks[i + 1] = tasks[i]
        if transitions[i] == 1:
            if tasks[i] == [1, 0]:
                tasks[i + 1] = [0, 1]
            if tasks[i] == [0, 1]:
                tasks[i + 1] = [1, 0]
    tasks = tasks[1:]

    # Determine correct response based on stimulus and task input
    correctResponse = np.sum(np.multiply(tasks, stimuli), axis=1)

    return tasks, stimuli, CSI, correctResponse


# Stability-Flexibility Model
def make_stab_flex(
    gain=3.0,
    leak=3.0,
    competition=2.0,
    lca_time_step_size=0.01,
    non_decision_time=0.2,
    automaticity=0.01,
    starting_value=0.0,
    threshold=0.1,
    ddm_noise=0.1,
    lca_noise=0.0,
    scale=0.2,
    ddm_time_step_size=0.01,
    rng_seed=None,
):

    GAIN = gain
    LEAK = leak
    COMP = competition
    AUTOMATICITY = automaticity  # Automaticity Weight

    STARTING_POINT = starting_value  # Starting Point
    THRESHOLD = threshold  # Threshold
    NOISE = ddm_noise  # Noise
    SCALE = scale  # Scales DDM inputs so threshold can be set to 1
    NON_DECISION_TIME = non_decision_time

    # Task Layer: [Color, Motion] {0, 1} Mutually Exclusive
    # Origin Node
    taskLayer = pnl.TransferMechanism(
        input_shapes=2,
        function=pnl.Linear(slope=1, intercept=0),
        output_ports=[pnl.RESULT],
        name="Task Input [I1, I2]",
    )

    # Stimulus Layer: [Color Stimulus, Motion Stimulus]
    # Origin Node
    stimulusInfo = pnl.TransferMechanism(
        input_shapes=2,
        function=pnl.Linear(slope=1, intercept=0),
        output_ports=[pnl.RESULT],
        name="Stimulus Input [S1, S2]",
    )

    # Cue-To-Stimulus Interval Layer
    # Origin Node
    cueInterval = pnl.TransferMechanism(
        input_shapes=1,
        function=pnl.Linear(slope=1, intercept=0),
        output_ports=[pnl.RESULT],
        name="Cue-Stimulus Interval",
    )

    # Correct Response Info
    # Origin Node
    correctResponseInfo = pnl.TransferMechanism(
        input_shapes=1,
        function=pnl.Linear(slope=1, intercept=0),
        output_ports=[pnl.RESULT],
        name="Correct Response Info",
    )

    # Control Module Layer: [Color Activation, Motion Activation]
    controlModule = pnl.LCAMechanism(
        input_shapes=2,
        function=pnl.Logistic(gain=GAIN),
        leak=LEAK,
        competition=COMP,
        self_excitation=0,
        noise=lca_noise,
        termination_measure=pnl.TimeScale.TRIAL,
        termination_threshold=1200,
        time_step_size=lca_time_step_size,
        name="Task Activations [Act1, Act2]",
    )

    # Control Mechanism Setting Cue-To-Stimulus Interval
    csiController = pnl.ControlMechanism(
        monitor_for_control=cueInterval,
        control_signals=[(pnl.TERMINATION_THRESHOLD, controlModule)],
        modulation=pnl.OVERRIDE,
    )

    # Hadamard product of controlModule and Stimulus Information
    nonAutomaticComponent = pnl.TransferMechanism(
        input_shapes=2,
        function=pnl.Linear(slope=1, intercept=0),
        input_ports=pnl.InputPort(combine=pnl.PRODUCT),
        output_ports=[pnl.RESULT],
        name="Non-Automatic Component [S1*Act1, S2*Act2]",
    )

    # Multiply Stimulus Input by the automaticity weight
    congruenceWeighting = pnl.TransferMechanism(
        input_shapes=2,
        function=pnl.Linear(slope=AUTOMATICITY, intercept=0),
        output_ports=[pnl.RESULT],
        name="Automaticity-weighted Stimulus Input [w*S1, w*S2]",
    )

    # Summation of nonAutomatic and Automatic Components
    ddmCombination = pnl.TransferMechanism(
        input_shapes=1,
        function=pnl.Linear(slope=1, intercept=0),
        input_ports=pnl.InputPort(combine=pnl.SUM),
        output_ports=[pnl.RESULT],
        name="Drift = (w*S1 + w*S2) + (S1*Act1 + S2*Act2)",
    )

    # Ensure upper boundary of DDM is always correct response by multiplying DDM input by correctResponseInfo
    ddmRecodeDrift = pnl.TransferMechanism(
        input_shapes=1,
        function=pnl.Linear(slope=1, intercept=0),
        input_ports=pnl.InputPort(combine=pnl.PRODUCT),
        output_ports=[pnl.RESULT],
        name="Recoded Drift = Drift * correctResponseInfo",
    )

    # Scale DDM inputs
    ddmInputScale = pnl.TransferMechanism(
        input_shapes=1,
        function=pnl.Linear(slope=SCALE, intercept=0),
        output_ports=[pnl.RESULT],
        name="Scaled DDM Input",
    )

    # Decision Module
    decisionMaker = pnl.DDM(
        function=pnl.DriftDiffusionIntegrator(
            starting_value=STARTING_POINT,
            threshold=THRESHOLD,
            noise=NOISE,
            time_step_size=ddm_time_step_size,
            non_decision_time=NON_DECISION_TIME,
        ),
        reset_stateful_function_when=pnl.AtTrialStart(),
        output_ports=[pnl.DECISION_OUTCOME, pnl.RESPONSE_TIME],
        name="DDM",
    )

    taskLayer.set_log_conditions([pnl.RESULT])
    stimulusInfo.set_log_conditions([pnl.RESULT])
    cueInterval.set_log_conditions([pnl.RESULT])
    correctResponseInfo.set_log_conditions([pnl.RESULT])
    controlModule.set_log_conditions([pnl.RESULT, "termination_threshold"])
    nonAutomaticComponent.set_log_conditions([pnl.RESULT])
    congruenceWeighting.set_log_conditions([pnl.RESULT])
    ddmCombination.set_log_conditions([pnl.RESULT])
    ddmRecodeDrift.set_log_conditions([pnl.RESULT])
    ddmInputScale.set_log_conditions([pnl.RESULT])
    decisionMaker.set_log_conditions([pnl.DECISION_OUTCOME, pnl.RESPONSE_TIME])

    # Composition Creation
    stabilityFlexibility = pnl.Composition()

    # Node Creation
    stabilityFlexibility.add_node(taskLayer)
    stabilityFlexibility.add_node(stimulusInfo)
    stabilityFlexibility.add_node(cueInterval)
    stabilityFlexibility.add_node(correctResponseInfo)
    stabilityFlexibility.add_node(controlModule)
    stabilityFlexibility.add_node(csiController)
    stabilityFlexibility.add_node(nonAutomaticComponent)
    stabilityFlexibility.add_node(congruenceWeighting)
    stabilityFlexibility.add_node(ddmCombination)
    stabilityFlexibility.add_node(ddmRecodeDrift)
    stabilityFlexibility.add_node(ddmInputScale)
    stabilityFlexibility.add_node(decisionMaker)

    # Projection Creation
    stabilityFlexibility.add_projection(sender=taskLayer, receiver=controlModule)
    stabilityFlexibility.add_projection(
        sender=controlModule, receiver=nonAutomaticComponent
    )
    stabilityFlexibility.add_projection(
        sender=stimulusInfo, receiver=nonAutomaticComponent
    )
    stabilityFlexibility.add_projection(
        sender=stimulusInfo, receiver=congruenceWeighting
    )
    stabilityFlexibility.add_projection(
        sender=nonAutomaticComponent, receiver=ddmCombination
    )
    stabilityFlexibility.add_projection(
        sender=congruenceWeighting, receiver=ddmCombination
    )
    stabilityFlexibility.add_projection(sender=ddmCombination, receiver=ddmRecodeDrift)
    stabilityFlexibility.add_projection(
        sender=correctResponseInfo, receiver=ddmRecodeDrift
    )
    stabilityFlexibility.add_projection(sender=ddmRecodeDrift, receiver=ddmInputScale)
    stabilityFlexibility.add_projection(sender=ddmInputScale, receiver=decisionMaker)

    # Hot-fix currently necessary to allow control module and DDM to execute in parallel in compiled mode
    # We need two gates in order to output both values (decision and response) from the ddm
    decisionGate = pnl.ProcessingMechanism(input_shapes=1, name="DECISION_GATE")
    stabilityFlexibility.add_node(decisionGate)

    responseGate = pnl.ProcessingMechanism(input_shapes=1, name="RESPONSE_GATE")
    stabilityFlexibility.add_node(responseGate)

    stabilityFlexibility.add_projection(
        sender=decisionMaker.output_ports[0], receiver=decisionGate
    )
    stabilityFlexibility.add_projection(
        sender=decisionMaker.output_ports[1], receiver=responseGate
    )

    # Sets scheduler conditions, so that the gates are not executed (and hence the composition doesn't finish) until decisionMaker is finished
    stabilityFlexibility.scheduler.add_condition(
        decisionGate, pnl.WhenFinished(decisionMaker)
    )
    stabilityFlexibility.scheduler.add_condition(
        responseGate, pnl.WhenFinished(decisionMaker)
    )

    return stabilityFlexibility

def get_node(comp, name):
    """
    Get the node from the composition with the given name. The name needs to match from the beginning, but it
    can have any numeric suffix after the name.
    """
    for node in comp.nodes:
        if node.name.startswith(name):
            return node
    return None


def make_input_dict(stab_flex_comp, taskTrain, stimulusTrain, cueTrain, correctResponse):
    inputs = {
        get_node(stab_flex_comp, "Task Input [I1, I2]"): [[np.array(v)] for v in taskTrain],
        get_node(stab_flex_comp, "Stimulus Input [S1, S2]"): [[np.array(v)] for v in stimulusTrain],
        get_node(stab_flex_comp, "Cue-Stimulus Interval"): [[np.array(v)] for v in cueTrain],
        get_node(stab_flex_comp, "Correct Response Info"): [[np.array(v)] for v in correctResponse]
    }

    return inputs

def run_stab_flex_cond(
        taskTrain,
        stimulusTrain,
        cueTrain,
        correctResponse,
        num_trials,
        **kwargs):
    """
    Create a stability flexibility composition and run it with the given parameters. Return the composition and the
    results as a pandas DataFrame. If any of the parameters are a list, then that parameter is assumed to be trial-wise
    and the length of the list should be the number of trials. A control mechanism will be added to the composition to
    override the parameter with the value from the input.
    """

    # Remove any parameters that are trial-wise from the kwargs, these values will be passed in as inputs
    # to the composition.
    cond_params = {name: value for name, value in kwargs.items()
                   if isinstance(value, list) or isinstance(value, np.ndarray)}

    # Remove the trial-wise parameters from the kwargs
    kwargs = {name: value for name, value in kwargs.items() if name not in cond_params}

    # Make a stability flexibility composition
    comp = make_stab_flex(**kwargs)

    inputs = make_input_dict(comp, taskTrain, stimulusTrain, cueTrain, correctResponse)

    # A dict to map keyword arg name to the corresponding mechanism in the composition
    param_map = {
        "gain": ("gain", comp.nodes["Task Activations [Act1, Act2]"]), # Gain
        "automaticity": ("slope", comp.nodes["Automaticity-weighted Stimulus Input [w*S1, w*S2]"]), # Automaticity
        "threshold": ("threshold", comp.nodes["DDM"]), # Threshold
        "non_decision_time": ("non_decision_time", comp.nodes["DECISION_GATE"]), # Non-decision time
    }
    # Go through the parameters and check if any are trial-wise, if so, add a control mechanism to override the value on
    # trial-by-trial basis with the value from the input.
    pec_mechs = {}
    for (name, value) in cond_params.items():

        if len(value) != num_trials:
            raise ValueError("Length of trial-wise parameter must be equal to the number of trials.")

        pec_mechs[name] = pnl.ControlMechanism(name=f"{name}_control",
                                               control_signals=param_map[name],
                                               modulation=pnl.OVERRIDE)
        comp.add_node(pec_mechs[name])
        inputs[pec_mechs[name]] = [[np.array([value[i]])] for i in range(num_trials)]

    comp.run(inputs, execution_mode=pnl.ExecutionMode.LLVMRun)

    df = pd.DataFrame(
        np.squeeze(np.array(comp.results))[:, 1:], columns=["decision", "response_time"]
    )
    df["decision"] = df["decision"].astype("category")

    # Add the trial-wise parameters to the DataFrame as well.
    for name in pec_mechs.keys():
        df[name] = cond_params[name]

    assert len(comp.input_ports) > 0

    return comp, df

@pytest.mark.llvm
def test_stab_flex_cond_fit():
    from psyneulink.core.globals.utilities import set_global_seed

    # # Let's make things reproducible
    pnl_seed = 42
    set_global_seed(pnl_seed)
    trial_seq_seed = 43

    # High-level parameters the impact performance of the test
    num_trials = 75
    time_step_size = 0.01
    num_estimates = 100

    sf_params = dict(
        gain=3.0,
        leak=3.0,
        competition=2.0,
        lca_time_step_size=time_step_size,
        non_decision_time=0.2,
        automaticity=0.01,
        starting_value=0.0,
        threshold=0.1,
        ddm_noise=0.1,
        lca_noise=0.0,
        scale=0.2,
        ddm_time_step_size=time_step_size,
    )

    # Generate some sample data to run the model on
    taskTrain, stimulusTrain, cueTrain, correctResponse = generate_trial_sequence(240, 0.5, seed=trial_seq_seed)
    taskTrain = taskTrain[0:num_trials]
    stimulusTrain = stimulusTrain[0:num_trials]
    cueTrain = cueTrain[0:num_trials]
    correctResponse = correctResponse[0:num_trials]

    # CSI is in terms of time steps, we need to scale by ten because original code
    # was set to run with timestep size of 0.001
    cueTrain = [c / 10.0 for c in cueTrain]

    # We will generate a dataset that comprises two different conditions. Each condition will have a different threshold.
    # Randomly select which trials will be in each condition uniformly.
    rng = np.random.default_rng(12345)
    threshold = rng.choice([0.3, 0.7], size=num_trials, replace=True)

    # Run
    _, data_to_fit = run_stab_flex_cond(
        taskTrain,
        stimulusTrain,
        cueTrain,
        correctResponse,
        num_trials,
        **{**sf_params, 'threshold': threshold}
    )

    # Turn our trial-wise threshold into a condition
    data_to_fit['condition'] = np.where(data_to_fit['threshold'] == 0.3, 'threshold=0.3', 'threshold=0.7')
    data_to_fit.drop(columns=['threshold'], inplace=True)

    # %%
    # Create a parameter estimation composition to fit the data we just generated and hopefully recover the
    # parameters of the composition.
    comp = make_stab_flex(**sf_params)

    controlModule = get_node(comp, "Task Activations [Act1, Act2]")
    congruenceWeighting = get_node(comp, "Automaticity-weighted Stimulus Input [w*S1, w*S2]")
    decisionMaker = get_node(comp, "DDM")
    decisionGate = get_node(comp, "DECISION_GATE")
    responseGate = get_node(comp, "RESPONSE_GATE")

    fit_parameters = {
        ("gain", controlModule): np.linspace(1.0, 10.0, 1000),  # Gain
        ("slope", congruenceWeighting): np.linspace(0.0, 0.1, 1000),  # Automaticity
        ("threshold", decisionMaker): np.linspace(0.01, 0.5, 1000),  # Threshold
        ("non_decision_time", decisionMaker): np.linspace(0.1, 0.4, 1000),  # Threshold
    }

    pec = pnl.ParameterEstimationComposition(
        name="pec",
        nodes=comp,
        parameters=fit_parameters,
        depends_on={("threshold", decisionMaker): 'condition'},
        outcome_variables=[
            decisionGate.output_ports[0],
            responseGate.output_ports[0],
        ],
        data=data_to_fit,
        optimization_function=PECOptimizationFunction(
            method=optuna.samplers.RandomSampler, max_iterations=10
        ),
        num_estimates=num_estimates,
        initial_seed=42,
    )

    pec.controller.parameters.comp_execution_mode.set("LLVM")
    pec.controller.function.parameters.save_values.set(True)

    inputs = make_input_dict(comp, taskTrain, stimulusTrain, cueTrain, correctResponse)

    pec.run(inputs=inputs)
    optimal_parameters = pec.optimized_parameter_values

    # These aren't the recovered parameters, we are doing too few trials and too few estimates to get the correct
    # results.
    expected_results = {
        'Task Activations [Act1, Act2]-1.gain': 2.3965500000000004,
        'Automaticity-weighted Stimulus Input [w*S1, w*S2]-1.slope': 0.0058000000000000005,
        'DDM-1.threshold[threshold=0.7]': 0.43483,
        'DDM-1.threshold[threshold=0.3]': 0.30449,
        'DDM-1.non_decision_time': 0.3124
    }

    for key, value in expected_results.items():
        np.testing.assert_allclose(optimal_parameters[key], value, rtol=1e-6)



 #
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *****************************************  OBJECTIVE FUNCTIONS *******************************************************
"""

* `Stability`
* `Energy`
* `Entropy`
* `Distance`

Functions that return a scalar evaluation of their input.

"""

import functools

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Callable

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.component import DefaultsFlexibility
from psyneulink.core.components.functions.function import EPSILON, FunctionError, Function_Base, get_matrix
from psyneulink.core.globals.keywords import \
    CORRELATION, COSINE, COSINE_SIMILARITY, CROSS_ENTROPY, \
    DEFAULT_VARIABLE, DIFFERENCE, DISTANCE_FUNCTION, DISTANCE_METRICS, DOT_PRODUCT, \
    ENERGY, ENTROPY, EUCLIDEAN, HOLLOW_MATRIX, MATRIX, MAX_ABS_DIFF, NORMALIZE, \
    NORMED_L0_SIMILARITY, OBJECTIVE_FUNCTION_TYPE, INPUT_SHAPES, STABILITY_FUNCTION
from psyneulink.core.globals.parameters import FunctionParameter, Parameter, check_user_specified, copy_parameter_value
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.utilities import DistanceMetricLiteral, safe_len, convert_to_np_array, convert_all_elements_to_np_array
from psyneulink.core.globals.utilities import is_iterable


__all__ = ['ObjectiveFunction', 'Stability', 'Distance', 'Energy', 'Entropy']


class ObjectiveFunction(Function_Base):
    """Abstract class of `Function` used for evaluating ports.
    """

    componentType = OBJECTIVE_FUNCTION_TYPE

    class Parameters(Function_Base.Parameters):
        """
            Attributes
            ----------

                metric
                    see `metric <ObjectiveFunction.metric>`

                    :default value: None
                    :type:

                normalize
                    see `normalize <ObjectiveFunction.normalize>`

                    :default value: False
                    :type: ``bool``
        """
        normalize = Parameter(False, stateful=False)
        metric = Parameter(None, stateful=False)


class Stability(ObjectiveFunction):
    """
    Stability(                                  \
        default_variable=None,                  \
        size=None,                              \
        matrix=HOLLOW_MATRIX,                   \
        metric=ENERGY                           \
        transfer_fct=None                       \
        normalize=False,                        \
        params=None,                            \
        owner=None,                             \
        prefs=None                              \
        )

    .. _StabilityFunction:

    Return the stability of `variable <Stability.variable>` based on a state transformation matrix.

    The value of `variable <Stability.variable>` is passed through the `matrix <Stability.matrix>`,
    transformed using the `transfer_fct <Stability.transfer_fct>` (if specified), and then compared with its initial
    value using the `distance metric <DistanceMetric>` specified by `metric <Stability.metric>`.  If `normalize
    <Stability.normalize>` is `True`, the result is normalized by the length of (number of elements in) `variable
    <Stability.variable>`.

    Arguments
    ---------

    variable : list or 1d array of numbers: Default class_defaults.variable
        specifies shape and default value of the array for which stability is calculated.

    input_shapes : int : None
        specifies length of the array over which stability is calculated;  can be used in place of default_value,
        in which case zeros are assigned as the value(s). An error is generated if both are specified but
        size != len(default_value).

    matrix : list, np.ndarray, or matrix keyword : default HOLLOW_MATRIX
        specifies the matrix of recurrent weights;  must be a square matrix with the same width as the
        length of `variable <Stability.variable>`.

    metric : keyword in DistanceMetrics : Default ENERGY
        specifies a `metric <DistanceMetrics>` from `DistanceMetrics` used to compute stability.

    transfer_fct : function or method : Default None
        specifies the function used to transform output of weight `matrix <Stability.matrix>`.

    normalize : bool : Default False
        specifies whether to normalize the stability value by the length of `variable <Stability.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).


    Attributes
    ----------

    variable : 1d array
        array for which stability is calculated.

    size : int
        length of array for which stability is calculated.

    matrix : list, np.ndarray, function keyword, or MappingProjection : default HOLLOW_MATRIX
        weight matrix from each element of `variable <Stability.variable>` to each other;  if a matrix other
        than HOLLOW_MATRIX is assigned, it is convolved with HOLLOW_MATRIX to eliminate self-connections from the
        stability calculation.

    metric : keyword in DistanceMetrics
        metric used to compute stability; must be a `DistanceMetrics` keyword. The `Distance` Function is used to
        compute the stability of `variable <Stability.variable>` with respect to its value after its transformation
        by `matrix <Stability.matrix>` and `transfer_fct <Stability.transfer_fct>`.

    transfer_fct : function or method
        function used to transform output of weight `matrix <Stability.matrix>` prior to computing stability.

    normalize : bool
        if `True`, result of stability calculation is normalized by the length of `variable <Stability.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).
    """

    componentName = STABILITY_FUNCTION

    class Parameters(ObjectiveFunction.Parameters):
        """
            Attributes
            ----------

                matrix
                    see `matrix <Stability.matrix>`

                    :default value: `HOLLOW_MATRIX`
                    :type: ``str``

                metric
                    see `metric <Stability.metric>`

                    :default value: `ENERGY`
                    :type: ``str``

                metric_fct
                    see `metric_fct <Stability.metric_fct>`

                    :default value: None
                    :type:

                transfer_fct
                    see `transfer_fct <Stability.transfer_fct>`

                    :default value: None
                    :type:
        """
        matrix = HOLLOW_MATRIX
        metric = Parameter(ENERGY, stateful=False)
        metric_fct = Parameter(None, stateful=False, loggable=False)
        transfer_fct = Parameter(None, stateful=False, loggable=False)
        normalize = FunctionParameter(False, function_name='metric_fct')

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 input_shapes=None,
                 matrix=None,
                 # metric:is_distance_metric=None,
                 metric: Optional[DistanceMetricLiteral] = None,
                 transfer_fct: Optional[Callable] = None,
                 normalize: Optional[bool] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):

        if input_shapes:
            if default_variable is None:
                default_variable = np.zeros(input_shapes)
            elif input_shapes != len(default_variable):
                raise FunctionError(f"Both {repr(DEFAULT_VARIABLE)} ({default_variable}) and {repr(INPUT_SHAPES)} ({input_shapes}) "
                                    f"are specified for {self.name} but are {INPUT_SHAPES}!=len({DEFAULT_VARIABLE}).")

        super().__init__(
            default_variable=default_variable,
            matrix=matrix,
            metric=metric,
            transfer_fct=transfer_fct,
            normalize=normalize,
            params=params,
            owner=owner,
            prefs=prefs,
        )

        # MODIFIED 6/12/19 NEW: [JDC]
        self._variable_shape_flexibility = DefaultsFlexibility.FLEXIBLE
        # MODIFIED 6/12/19 END

    def _validate_variable(self, variable, context=None):
        """Validates that variable is 1d array
        """
        if len(np.atleast_2d(variable)) != 1:
            raise FunctionError("Variable for {} must contain a single array or list of numbers".format(self.name))
        return variable

    def _validate_params(self, variable, request_set, target_set=None, context=None):
        """Validate matrix param

        `matrix <Stability.matrix>` argument must be one of the following
            - 2d list, np.ndarray
            - ParameterPort for one of the above
            - MappingProjection with a parameterPorts[MATRIX] for one of the above

        Parse matrix specification to insure it resolves to a square matrix
        (but leave in the form in which it was specified so that, if it is a ParameterPort or MappingProjection,
         its current value can be accessed at runtime (i.e., it can be used as a "pointer")
        """

        # Validate matrix specification
        # (str can be automatically transformed to variable shape)
        if MATRIX in target_set and not isinstance(target_set[MATRIX], str):

            from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
            from psyneulink.core.components.ports.parameterport import ParameterPort

            matrix = target_set[MATRIX]

            if isinstance(matrix, MappingProjection):
                try:
                    matrix = matrix._parameter_ports[MATRIX].value
                    param_type_string = "MappingProjection's ParameterPort"
                except KeyError:
                    raise FunctionError("The MappingProjection specified for the {} arg of {} ({}) must have a {} "
                                        "ParameterPort that has been assigned a 2d array or matrix".
                                        format(MATRIX, self.name, matrix.shape, MATRIX))

            elif isinstance(matrix, ParameterPort):
                try:
                    matrix = matrix.value
                    param_type_string = "ParameterPort"
                except KeyError:
                    raise FunctionError("The value of the {} parameterPort specified for the {} arg of {} ({}) "
                                        "must be a 2d array or matrix".
                                        format(MATRIX, MATRIX, self.name, matrix.shape))

            else:
                param_type_string = "array or matrix"

            matrix = np.array(matrix)
            if matrix.ndim != 2:
                raise FunctionError("The value of the {} specified for the {} arg of {} ({}) "
                                    "must be a 2d array or matrix".
                                    format(param_type_string, MATRIX, self.name, matrix))
            rows = matrix.shape[0]
            cols = matrix.shape[1]

            # this mirrors the transformation in _function
            # it is a hack, and a general solution should be found
            squeezed = np.array(self.defaults.variable)
            if squeezed.ndim > 1:
                squeezed = np.squeeze(squeezed)

            size = safe_len(squeezed)

            if rows != size:
                raise FunctionError("The value of the {} specified for the {} arg of {} is the wrong size;"
                                    "it is {}x{}, but must be square matrix of size {}".
                                    format(param_type_string, MATRIX, self.name, rows, cols, size))

            if rows != cols:
                raise FunctionError("The value of the {} specified for the {} arg of {} ({}) "
                                    "must be a square matrix".
                                    format(param_type_string, MATRIX, self.name, matrix))

        super()._validate_params(request_set=request_set,
                                 target_set=target_set,
                                 context=context)

    def _parse_metric_fct_variable(self, variable):
        return convert_to_np_array([variable, variable])

    def _instantiate_attributes_before_function(self, function=None, context=None):
        """Instantiate matrix

        Specified matrix is convolved with HOLLOW_MATRIX
            to eliminate the diagonal (self-connections) from the calculation.
        The `Distance` Function is used for all calculations except ENERGY (which is not really a distance metric).
        If ENTROPY is specified as the metric, convert to CROSS_ENTROPY for use with the Distance Function.
        :param function:

        """
        from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
        from psyneulink.core.components.ports.parameterport import ParameterPort

        # this mirrors the transformation in _function
        # it is a hack, and a general solution should be found
        squeezed = np.array(self.defaults.variable)
        if squeezed.ndim > 1:
            squeezed = np.squeeze(squeezed)

        size = safe_len(squeezed)

        matrix = self.parameters.matrix._get(context)

        if isinstance(matrix, MappingProjection):
            matrix = matrix._parameter_ports[MATRIX]
        # elif isinstance(matrix, ParameterPort):
        #     pass
        else:
            matrix = get_matrix(matrix, size, size)

        self.parameters.matrix._set(matrix, context)

        self._hollow_matrix = get_matrix(HOLLOW_MATRIX, size, size)

        default_variable = [self.defaults.variable,
                            self.defaults.variable]

        if self.metric == ENTROPY:
            self.metric_fct = Distance(default_variable=default_variable,
                                       metric=CROSS_ENTROPY,
                                       normalize=self.parameters.normalize.default_value)
        elif self.metric in DISTANCE_METRICS._set():
            self.metric_fct = Distance(default_variable=default_variable,
                                       metric=self.metric,
                                       normalize=self.parameters.normalize.default_value)
        else:
            assert False, "Unknown metric"

        #FIXME: This is a hack to make sure metric-fct param is set
        self.parameters.metric_fct.set(self.metric_fct)

    def _update_default_variable(self, new_default_variable, context):
        from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
        from psyneulink.core.components.ports.parameterport import ParameterPort

        # this mirrors the transformation in _function
        # it is a hack, and a general solution should be found
        new_default_variable = convert_all_elements_to_np_array(new_default_variable)
        size = safe_len(np.squeeze(new_default_variable))
        matrix = self.parameters.matrix._get(context)

        if isinstance(matrix, MappingProjection):
            matrix = matrix._parameter_ports[MATRIX]
        elif isinstance(matrix, ParameterPort):
            pass
        else:
            matrix = get_matrix(copy_parameter_value(self.defaults.matrix), size, size)

        self.parameters.matrix._set(matrix, context)

        self._hollow_matrix = get_matrix(HOLLOW_MATRIX, size, size)

        super()._update_default_variable(new_default_variable, context)

    def _gen_llvm_function_body(self, ctx, builder, params, state, arg_in, arg_out, *, tags:frozenset):
        # Dot product
        dot_out = builder.alloca(arg_in.type.pointee)
        matrix = ctx.get_param_or_state_ptr(builder, self, MATRIX, param_struct_ptr=params, state_struct_ptr=state)

        # Convert array pointer to pointer to the fist element
        matrix = builder.gep(matrix, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_in = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_out = builder.gep(dot_out, [ctx.int32_ty(0), ctx.int32_ty(0)])

        input_length = ctx.int32_ty(arg_in.type.pointee.count)
        output_length = ctx.int32_ty(arg_in.type.pointee.count)
        builtin = ctx.import_llvm_function("__pnl_builtin_vxm")
        builder.call(builtin, [vec_in, matrix, input_length, output_length, vec_out])

        # Prepare metric function
        metric_fun = ctx.import_llvm_function(self.metric_fct)
        metric_in = builder.alloca(metric_fun.args[2].type.pointee)

        # Transfer Function if configured
        if self.transfer_fct is not None:
            #FIXME: implement this
            assert False, "Support for transfer functions is not implemented"
        else:
            # Check that transfer_fct is absent from the compiled parameter
            # structure or represented by an empty structure
            assert "transfer_fct" not in self.llvm_param_ids or ctx.get_param_or_state_ptr(builder, self, "transfer_fct", param_struct_ptr=params).type.pointee.elements == ()

            trans_out = builder.gep(metric_in, [ctx.int32_ty(0), ctx.int32_ty(1)])
            builder.store(builder.load(dot_out), trans_out)

        # Copy original variable
        builder.store(builder.load(arg_in), builder.gep(metric_in, [ctx.int32_ty(0), ctx.int32_ty(0)]))

        # Distance Function
        metric_params, metric_state = ctx.get_param_or_state_ptr(builder, self, "metric_fct", param_struct_ptr=params, state_struct_ptr=state)
        metric_out = arg_out
        builder.call(metric_fun, [metric_params, metric_state, metric_in, metric_out])
        return builder

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """Calculate the stability of `variable <Stability.variable>`.

        Compare the value of `variable <Stability.variable>` with its value after transformation by
        `matrix <Stability.matrix>` and `transfer_fct <Stability.transfer_fct>` (if specified), using the specified
        `metric <Stability.metric>`.  If `normalize <Stability.normalize>` is `True`, the result is divided
        by the length of `variable <Stability.variable>`.

        Returns
        -------

        stability : scalar

        """

        # MODIFIED 6/12/19 NEW: [JDC]
        variable = np.array(variable)
        if variable.ndim > 1:
            variable = np.squeeze(variable)
        # MODIFIED 6/12/19 END

        matrix = self._get_current_parameter_value(MATRIX, context)
        if matrix is None:
            matrix = self.matrix

        current = variable

        transformed = np.dot(matrix * self._hollow_matrix, variable)
        if self.transfer_fct is not None:
            transformed = self.transfer_fct(transformed)

        result = self.metric_fct(variable=[current, transformed], context=context)

        return self.convert_output_type(result)


class Energy(Stability):
    """
    Energy(                           \
        default_variable=None,        \
        size=None,                    \
        matrix=INVERSE_HOLLOW_MATRIX, \
        transfer_fct=None             \
        normalize=False,              \
        params=None,                  \
        owner=None,                   \
        prefs=None                    \
        )

    .. _EnergyFunction:

    Subclass of `Stability` Function that returns the energy of an array.

    Arguments
    ---------

    variable : list or 1d array of numbers: Default class_defaults.variable
        specifies shape and default value of the array for which energy is calculated.

    input_shapes : int : None
        specifies length of the array over which energy is calculated;  can be used in place of default_value,
        in which case zeros are assigned as the value(s). An error is generated if both are specified but
        size != len(default_value).

    matrix : list, np.ndarray, or matrix keyword : default INVERSE_HOLLOW_MATRIX
        specifies the matrix of recurrent weights;  must be a square matrix with the same width as the
        length of `variable <Stability.variable>`.

    transfer_fct : function or method : Default None
        specifies the function used to transform output of `matrix <Stability.matrix>` prior to the energy calculation
        (see `Stablility <Stability>` for explanation).

    normalize : bool : Default False
        specifies whether to normalize the energy value by the length of `variable <Stability.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).


    Attributes
    ----------

    variable : 1d array
        array for which energy is calculated.

    size : int
        length of array for which energy is calculated.

    matrix : list, np.ndarray, or matrix keyword
        weight matrix from each element of `variable <Energy.variable>` to each other;  if a matrix other
        than INVERSE_HOLLOW_MATRIX is assigned, it is convolved with HOLLOW_MATRIX to eliminate self-connections from
        the energy calculation.

    transfer_fct : function or method
        function used to transform output of `matrix <Stability.matrix>` prior to computing energy
        (see `Stability` for explanation).

    normalize : bool
        if `True`, result of energy calculation is normalized by the length of `variable <Energy.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).
    """

    @check_user_specified
    def __init__(self,
                 default_variable=None,
                 input_shapes=None,
                 normalize:bool=None,
                 transfer_fct=None,
                 matrix=None,
                 params=None,
                 owner=None,
                 prefs=None):

        super().__init__(
            default_variable=default_variable,
            input_shapes=input_shapes,
            metric=ENERGY,
            matrix=matrix,
            transfer_fct=transfer_fct,
            normalize=normalize,
            params=params,
            owner=owner,
            prefs=prefs)


class Entropy(Stability):
    """
    Entropy(                          \
        default_variable=None,        \
        input_shapes=None,            \
        matrix=INVERSE_HOLLOW_MATRIX, \
        transfer_fct=None             \
        normalize=False,              \
        params=None,                  \
        owner=None,                   \
        prefs=None                    \
        )

    .. _EntropyFunction:

    Subclass of `Stability` Function that returns the entropy of an array.

    Arguments
    ---------

    variable : list or 1d array of numbers: Default class_defaults.variable
        specifies shape and default value of the array for which entropy is calculated.

    input_shapes : int : None
        specifies length of the array over which entropy is calculated;  can be used in place of default_value,
        in which case zeros are assigned as the value(s). An error is generated if both are specified but
        input_shapes != len(default_value).

    matrix : list, np.ndarray, or matrix keyword : default INVERSE_HOLLOW_MATRIX
        specifies the matrix of recurrent weights;  must be a square matrix with the same width as the
        length of `variable <Stability.variable>`.

    transfer_fct : function or method : Default None
        specifies the function used to transform output of `matrix <Stability.matrix>` prior to the entropy calculation
        (see `Stability` for explanation).

    normalize : bool : Default False
        specifies whether to normalize the entropy value by the length of `variable <Stability.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).


    Attributes
    ----------

    variable : 1d array
        array for which entropy is calculated.

    input_shapes : int
        length of array for which energy is calculated.

    matrix : list, np.ndarray, or matrix keyword
        weight matrix from each element of `variable <Entropy.variable>` to each other;  if a matrix other
        than INVERSE_HOLLOW_MATRIX is assigned, it is convolved with HOLLOW_MATRIX to eliminate self-connections from
        the entropy calculation.

    transfer_fct : function or method
        function used to transform output of `matrix <Stability.matrix>` prior to computing entropy
        (see `Stability` for explanation).

    normalize : bool
        if `True`, result of entropy calculation is normalized by the length of `variable <Entropy.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).
    """

    @check_user_specified
    def __init__(self,
                 default_variable=None,
                 input_shapes=None,
                 normalize:bool=None,
                 matrix=None,
                 transfer_fct=None,
                 params=None,
                 owner=None,
                 prefs=None):

        super().__init__(
            default_variable=default_variable,
            input_shapes=input_shapes,
            metric=ENTROPY,
            matrix=matrix,
            transfer_fct=transfer_fct,
            normalize=normalize,
            params=params,
            owner=owner,
            prefs=prefs)


class Distance(ObjectiveFunction):
    """
    Distance(                                    \
       default_variable=None,                    \
       metric=EUCLIDEAN                          \
       normalize=False,                          \
       params=None,                              \
       owner=None,                               \
       prefs=None                                \
       )

    .. _Distance:

    Return the distance between the vectors in the two items of `variable <Distance.variable>` using the `distance
    metric <DistanceMetrics>` specified in the `metric <Stability.metric>` attribute.  If `normalize
    <Distance.normalize>` is `True`, the result is normalized by the length of (number of elements in) `variable
    <Stability.variable>`.

    Arguments
    ---------

    variable : 2d array with two items : Default class_defaults.variable
        the arrays between which the distance is calculated.

    metric : keyword in DistancesMetrics : Default EUCLIDEAN
        specifies a `distance metric <DistanceMetrics>` used to compute the distance between the two items in `variable
        <Distance.variable>`.

    normalize : bool : Default False
        specifies whether to normalize the distance by the length of `variable <Distance.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : 2d array with two items
        contains the arrays between which the distance is calculated.

    metric : keyword in DistanceMetrics
        determines the `metric <DistanceMetrics>` used to compute the distance between the two items in `variable
        <Distance.variable>`.

    normalize : bool
        determines whether the distance is normalized by the length of `variable <Distance.variable>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).
    """

    componentName = DISTANCE_FUNCTION

    class Parameters(ObjectiveFunction.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <Distance.variable>`

                    :default value: numpy.array([[0], [0]])
                    :type: ``numpy.ndarray``
                    :read only: True

                metric
                    see `metric <Distance.metric>`

                    :default value: `DIFFERENCE`
                    :type: ``str``
        """
        variable = Parameter(np.array([[0], [0]]), read_only=True, pnl_internal=True, constructor_argument='default_variable')
        metric = Parameter(DIFFERENCE, stateful=False)

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 metric: Optional[DistanceMetricLiteral] = None,
                 normalize: Optional[bool] = None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):
        super().__init__(
            default_variable=default_variable,
            metric=metric,
            normalize=normalize,
            params=params,
            owner=owner,
            prefs=prefs,
        )

    def _validate_params(self, request_set, target_set=None, variable=None, context=None):
        """Validate that variable has two items of equal length

        """
        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        err_two_items = FunctionError("variable for {} ({}) must have two items".format(self.name, variable))

        try:
            if len(variable) != 2:
                raise err_two_items
        except TypeError:
            raise err_two_items

        try:
            if len(variable[0]) != len(variable[1]):
                raise FunctionError(
                    "The lengths of the items in the variable for {0} ({1},{2}) must be equal".format(
                        self.name,
                        variable[0],
                        variable[1]
                    )
                )
        except TypeError:
            if is_iterable(variable[0]) ^ is_iterable(variable[1]):
                raise FunctionError(
                    "The lengths of the items in the variable for {0} ({1},{2}) must be equal".format(
                        self.name,
                        variable[0],
                        variable[1]
                    )
                )

    def cosine(v1, v2):
        numer = np.sum(v1 * v2)
        denom = np.sqrt(np.sum(v1 ** 2)) * np.sqrt(np.sum(v2 ** 2)) or EPSILON
        return numer / denom

    def correlation(v1, v2):
        v1_norm = v1 - np.mean(v1)
        v2_norm = v2 - np.mean(v2)
        denom = np.sqrt(np.sum(v1_norm ** 2) * np.sum(v2_norm ** 2)) or EPSILON
        return np.sum(v1_norm * v2_norm) / denom

    def __gen_llvm_sum_difference(self, builder, index, ctx, v1, v2, acc):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        sub = builder.fsub(val1, val2)
        abs_f = ctx.get_builtin("fabs", [sub.type])
        abs_val = builder.call(abs_f, [sub])
        acc_val = builder.load(acc)
        new_acc = builder.fadd(acc_val, abs_val)
        builder.store(new_acc, acc)

    def __gen_llvm_sum_diff_squares(self, builder, index, ctx, v1, v2, acc):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        sub = builder.fsub(val1, val2)
        sqr = builder.fmul(sub, sub)
        acc_val = builder.load(acc)
        new_acc = builder.fadd(acc_val, sqr)
        builder.store(new_acc, acc)

    def __gen_llvm_cross_entropy(self, builder, index, ctx, v1, v2, acc):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        log_f = ctx.get_builtin("log", [ctx.float_ty])
        log = builder.call(log_f, [val2])
        prod = builder.fmul(val1, log)

        acc_val = builder.load(acc)
        new_acc = builder.fsub(acc_val, prod)
        builder.store(new_acc, acc)

    def __gen_llvm_sum_product(self, builder, index, ctx, v1, v2, acc):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        prod = builder.fmul(val1, val2)

        acc_val = builder.load(acc)
        new_acc = builder.fadd(acc_val, prod)
        builder.store(new_acc, acc)

    def __gen_llvm_cosine(self, builder, index, ctx, v1, v2, numer_acc, denom1_acc, denom2_acc):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        # Numerator
        numer = builder.load(numer_acc)
        val = builder.fmul(val1, val2)
        numer = builder.fadd(numer, val)
        builder.store(numer, numer_acc)

        # Denominator1
        denom = builder.load(denom1_acc)
        val = builder.fmul(val1, val1)
        denom = builder.fadd(denom, val)
        builder.store(denom, denom1_acc)

        # Denominator2
        denom = builder.load(denom2_acc)
        val = builder.fmul(val2, val2)
        denom = builder.fadd(denom, val)
        builder.store(denom, denom2_acc)

    def __gen_llvm_max_diff(self, builder, index, ctx, v1, v2, max_diff_ptr):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        # Get the difference
        diff = builder.fsub(val1, val2)

        # Get absolute value
        fabs = ctx.get_builtin("fabs", [ctx.float_ty])
        diff = builder.call(fabs, [diff])

        # Get maximum
        old_max = builder.load(max_diff_ptr)
        fmax = ctx.get_builtin("maxnum", [ctx.float_ty])
        max_diff = builder.call(fmax, [diff, old_max])

        builder.store(max_diff, max_diff_ptr)

    def __gen_llvm_pearson(self, builder, index, ctx, v1, v2, acc_x, acc_y, acc_xy, acc_x2, acc_y2):
        ptr1 = builder.gep(v1, [index])
        ptr2 = builder.gep(v2, [index])
        val1 = builder.load(ptr1)
        val2 = builder.load(ptr2)

        # Sum X
        acc_x_val = builder.load(acc_x)
        acc_x_val = builder.fadd(acc_x_val, val1)
        builder.store(acc_x_val, acc_x)

        # Sum Y
        acc_y_val = builder.load(acc_y)
        acc_y_val = builder.fadd(acc_y_val, val2)
        builder.store(acc_y_val, acc_y)

        # Sum XY
        acc_xy_val = builder.load(acc_xy)
        xy = builder.fmul(val1, val2)
        acc_xy_val = builder.fadd(acc_xy_val, xy)
        builder.store(acc_xy_val, acc_xy)

        # Sum X2
        acc_x2_val = builder.load(acc_x2)
        x2 = builder.fmul(val1, val1)
        acc_x2_val = builder.fadd(acc_x2_val, x2)
        builder.store(acc_x2_val, acc_x2)

        # Sum Y2
        acc_y2_val = builder.load(acc_y2)
        y2 = builder.fmul(val2, val2)
        acc_y2_val = builder.fadd(acc_y2_val, y2)
        builder.store(acc_y2_val, acc_y2)

    def _gen_llvm_function_body(self, ctx, builder, params, _, arg_in, arg_out, *, tags:frozenset):
        assert isinstance(arg_in.type.pointee, pnlvm.ir.ArrayType)
        assert isinstance(arg_in.type.pointee.element, pnlvm.ir.ArrayType)
        assert arg_in.type.pointee.count == 2

        v1 = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(0)])
        v2 = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(1), ctx.int32_ty(0)])

        acc_ptr = builder.alloca(ctx.float_ty)
        builder.store(acc_ptr.type.pointee(-0.0), acc_ptr)

        kwargs = {"ctx": ctx, "v1": v1, "v2": v2, "acc": acc_ptr}
        if self.metric == DIFFERENCE or self.metric == NORMED_L0_SIMILARITY:
            inner = functools.partial(self.__gen_llvm_sum_difference, **kwargs)
        elif self.metric == EUCLIDEAN:
            inner = functools.partial(self.__gen_llvm_sum_diff_squares, **kwargs)
        elif self.metric == ENERGY or self.metric == DOT_PRODUCT:
            inner = functools.partial(self.__gen_llvm_sum_product, **kwargs)
        elif self.metric == CROSS_ENTROPY:
            inner = functools.partial(self.__gen_llvm_cross_entropy, **kwargs)
        elif self.metric in {COSINE, COSINE_SIMILARITY}:
            del kwargs['acc']
            numer_acc = builder.alloca(ctx.float_ty)
            denom1_acc = builder.alloca(ctx.float_ty)
            denom2_acc = builder.alloca(ctx.float_ty)
            for loc in numer_acc, denom1_acc, denom2_acc:
                builder.store(loc.type.pointee(-0.0), loc)

            kwargs['numer_acc'] = numer_acc
            kwargs['denom1_acc'] = denom1_acc
            kwargs['denom2_acc'] = denom2_acc
            inner = functools.partial(self.__gen_llvm_cosine, **kwargs)
        elif self.metric == MAX_ABS_DIFF:
            del kwargs['acc']
            max_diff_ptr = builder.alloca(ctx.float_ty)
            builder.store(max_diff_ptr.type.pointee(float("NaN")), max_diff_ptr)
            kwargs['max_diff_ptr'] = max_diff_ptr
            inner = functools.partial(self.__gen_llvm_max_diff, **kwargs)
        elif self.metric == CORRELATION:
            acc_x_ptr = builder.alloca(ctx.float_ty)
            acc_y_ptr = builder.alloca(ctx.float_ty)
            acc_xy_ptr = builder.alloca(ctx.float_ty)
            acc_x2_ptr = builder.alloca(ctx.float_ty)
            acc_y2_ptr = builder.alloca(ctx.float_ty)
            for loc in [acc_x_ptr, acc_y_ptr, acc_xy_ptr, acc_x2_ptr, acc_y2_ptr]:
                builder.store(loc.type.pointee(-0.0), loc)
            del kwargs['acc']
            kwargs['acc_x'] = acc_x_ptr
            kwargs['acc_y'] = acc_y_ptr
            kwargs['acc_xy'] = acc_xy_ptr
            kwargs['acc_x2'] = acc_x2_ptr
            kwargs['acc_y2'] = acc_y2_ptr
            inner = functools.partial(self.__gen_llvm_pearson, **kwargs)
        else:
            raise RuntimeError('Unsupported metric')

        input_length = arg_in.type.pointee.element.count
        vector_length = ctx.int32_ty(input_length)
        with pnlvm.helpers.for_loop_zero_inc(builder, vector_length, self.metric) as args:
            inner(*args)

        sqrt = ctx.get_builtin("sqrt", [ctx.float_ty])
        fabs = ctx.get_builtin("fabs", [ctx.float_ty])
        ret = builder.load(acc_ptr)
        if self.metric == NORMED_L0_SIMILARITY:
            ret = builder.fdiv(ret, ret.type(4))
            ret = builder.fsub(ret.type(1), ret)
        elif self.metric == DOT_PRODUCT:
            pass # the dot product has already been computed above by __gen_llvm_sum_product
        elif self.metric == ENERGY:
            ret = builder.fmul(ret, ret.type(-0.5))
        elif self.metric == EUCLIDEAN:
            ret = builder.call(sqrt, [ret])
        elif self.metric == MAX_ABS_DIFF:
            ret = builder.load(max_diff_ptr)
        elif self.metric in {COSINE, COSINE_SIMILARITY}:
            numer = builder.load(numer_acc)
            denom1 = builder.load(denom1_acc)
            denom1 = builder.call(sqrt, [denom1])
            denom2 = builder.load(denom2_acc)
            denom2 = builder.call(sqrt, [denom2])
            denom = builder.fmul(denom1, denom2)

            ret = builder.fdiv(numer, denom)
            ret = builder.call(fabs, [ret])
            ret = builder.fsub(ret.type(1), ret)

        elif self.metric == CORRELATION:
            n = ctx.float_ty(input_length)
            acc_xy = builder.load(acc_xy_ptr)
            acc_x = builder.load(acc_x_ptr)
            acc_y = builder.load(acc_y_ptr)
            acc_x2 = builder.load(acc_x2_ptr)
            acc_y2 = builder.load(acc_y2_ptr)

            # We'll need mean_x,y below
            mean_x = builder.fdiv(acc_x, n)
            mean_y = builder.fdiv(acc_y, n)

            # Numerator: sum((x - mean(x))*(y - mean(y)) =
            # sum(x*y - x*mean(y) - y*mean(x) + mean(x)*mean(y)) =
            # sum(x*y) - sum(x)*mean(y) - sum(y)*mean(x) + mean(x)*mean(y)*n
            b = builder.fmul(acc_x, mean_y)
            c = builder.fmul(acc_y, mean_x)
            d = builder.fmul(mean_x, mean_y)
            d = builder.fmul(d, n)

            numerator = builder.fsub(acc_xy, b)
            numerator = builder.fsub(numerator, c)
            numerator = builder.fadd(numerator, d)

            # Denominator: sqrt(D_X * D_Y)
            # D_X = sum((x - mean(x))^2) = sum(x^2 - 2*x*mean(x) + mean(x)^2) =
            # sum(x^2) - 2 * sum(x) * mean(x) + n * mean(x)^2
            dxb = builder.fmul(acc_x, mean_x)
            dxb = builder.fadd(dxb, dxb)        # *2
            dxc = builder.fmul(mean_x, mean_x)  # ^2
            dxc = builder.fmul(dxc, n)

            dx = builder.fsub(acc_x2, dxb)
            dx = builder.fadd(dx, dxc)

            # Similarly for y
            dyb = builder.fmul(acc_y, mean_y)
            dyb = builder.fadd(dyb, dyb)        # *2
            dyc = builder.fmul(mean_y, mean_y)  # ^2
            dyc = builder.fmul(dyc, n)

            dy = builder.fsub(acc_y2, dyb)
            dy = builder.fadd(dy, dyc)

            # Denominator: sqrt(D_X * D_Y)
            denominator = builder.fmul(dx, dy)
            denominator = builder.call(sqrt, [denominator])

            corr = builder.fdiv(numerator, denominator)

            # ret =  1 - abs(corr)
            ret = builder.call(fabs, [corr])
            ret = builder.fsub(ret.type(1), ret)

        if arg_out.type.pointee != ret.type:
            # Some instances use 2d output values
            arg_out = builder.gep(arg_out, [ctx.int32_ty(0), ctx.int32_ty(0),
                                            ctx.int32_ty(0)])

        normalize_ptr = ctx.get_param_or_state_ptr(builder, self, NORMALIZE, param_struct_ptr=params)
        normalize = builder.load(normalize_ptr)
        normalize_b = builder.fcmp_ordered("!=", normalize, normalize.type(0))

        # MAX_ABS_DIFF, CORRELATION, and COSINE/COSINE_SIMILARITY ignore normalization
        allow_normalize_b = normalize_b.type(self.metric not in {MAX_ABS_DIFF, CORRELATION, COSINE, COSINE_SIMILARITY})
        normalize_b = builder.and_(normalize_b, allow_normalize_b)
        with builder.if_else(normalize_b) as (then, otherwise):
            with then:
                norm_factor = input_length ** 2 if self.metric == ENERGY else input_length
                normalized = builder.fdiv(ret, ret.type(norm_factor), name="normalized")
                builder.store(normalized, arg_out)
            with otherwise:
                builder.store(ret, arg_out)

        return builder

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """Calculate the distance between the two vectors in `variable <Stability.variable>`.

        Use the `distance metric <DistanceMetrics>` specified in `metric <Distance.metric>` to calculate the distance.
        If `normalize <Distance.normalize>` is `True`, the result is divided by the length of `variable
        <Distance.variable>`.

        Returns
        -------

        distance : scalar

        """

        try:
            v1 = np.hstack(variable[0])
        except TypeError:
            v1 = variable[0]

        try:
            v2 = np.hstack(variable[1])
        except TypeError:
            v2 = variable[1]

        # Maximum of Hadamard (elementwise) difference of v1 and v2
        if self.metric == MAX_ABS_DIFF:
            result = np.max(np.fabs(v1 - v2))

        # Simple Hadamard (elementwise) difference of v1 and v2
        elif self.metric == DIFFERENCE:
            result = np.sum(np.fabs(v1 - v2))

        # Similarity (used specifically for testing Compilation of Predator-Prey Model)
        elif self.metric == NORMED_L0_SIMILARITY:
            result = 1.0 - np.sum(np.abs(v1 - v2)) / 4.0

        # Simple dot product of v1 and v2
        elif self.metric == DOT_PRODUCT:
            result = np.dot(v1, v2)

        # Euclidean distance between v1 and v2
        elif self.metric == EUCLIDEAN:
            result = np.linalg.norm(v2 - v1)

        # Cosine similarity of v1 and v2
        elif self.metric in {COSINE, COSINE_SIMILARITY}:
            result = 1.0 - np.fabs(Distance.cosine(v1, v2))

        # Correlation of v1 and v2
        elif self.metric == CORRELATION:
            # result = np.correlate(v1, v2)
            result = 1.0 - np.fabs(Distance.correlation(v1, v2))

        # FIX: IMPLEMENT VERSION THAT DIRECTLY COMPUTES THE LUCE RATIO
        # Cross-entropy of v1 and v2
        elif self.metric == CROSS_ENTROPY:
            # FIX: VALIDATE THAT ALL ELEMENTS OF V1 AND V2 ARE 0 TO 1
            if not self.is_initializing:
                v1 = np.where(v1 == 0, EPSILON, v1)
                v2 = np.where(v2 == 0, EPSILON, v2)
            # MODIFIED CW 3/20/18: avoid divide by zero error by plugging in two zeros
            # FIX: unsure about desired behavior when v2 = 0 and v1 != 0
            # JDC: returns [inf]; leave, and let it generate a warning or error message for user
            both_zero = np.logical_and(v1 == 0, v2 == 0)
            result = -np.sum(v1 * np.where(both_zero, 0.0, np.log(v2, where=np.logical_not(both_zero))))

        # Energy
        elif self.metric == ENERGY:
            result = -np.sum(v1 * v2) / 2.0

        else:
            assert False, '{} not a recognized metric in {}'.format(self.metric, self.__class__.__name__)

        if self.normalize and self.metric not in {MAX_ABS_DIFF, CORRELATION, COSINE, COSINE_SIMILARITY}:
            if self.metric == ENERGY:
                result /= len(v1) ** 2.0
            else:
                result /= len(v1)

        return self.convert_output_type(result)
